from typing import List

import attr

__all__ = [
    "BlockInfo",
    "Block",
    "Header",
    "Evidence",
    "BlockID",
    "Parts",
    "Version",
    "LastCommit",
    "Signature",
]


@attr.s
class Evidence:
    evidence: List[str] = attr.ib(converter=list)


@attr.s
class Version:
    block: str = attr.ib()
    app: str = attr.ib()


@attr.s
class Parts:
    total: str = attr.ib()
    hash: str = attr.ib()


@attr.s
class BlockID:
    hash: str = attr.ib()
    part_set_header: Parts = attr.ib()


@attr.s
class Header:
    version: Version = attr.ib()
    chain_id: str = attr.ib()
    height: str = attr.ib()
    time: str = attr.ib()
    last_block_id: BlockID = attr.ib()
    last_commit_hash: str = attr.ib()
    data_hash: str = attr.ib()
    validators_hash: str = attr.ib()
    next_validators_hash: str = attr.ib()
    consensus_hash: str = attr.ib()
    app_hash: str = attr.ib()
    last_results_hash: str = attr.ib()
    evidence_hash: str = attr.ib()
    proposer_address: str = attr.ib()


@attr.s
class Signature:
    block_id_flag: int = attr.ib(converter=int)
    validator_address: str = attr.ib()
    timestamp: str = attr.ib()
    signature: str = attr.ib()


@attr.s
class LastCommit:
    height: str = attr.ib()
    round: int = attr.ib(converter=int)
    block_id: BlockID = attr.ib()
    signatures: List[Signature] = attr.ib()


@attr.s
class Block:
    header: Header = attr.ib()
    data: List[str] = attr.ib()
    evidence: Evidence = attr.ib()
    last_commit: LastCommit = attr.ib()


@attr.s
class BlockInfo:
    block_id: BlockID = attr.ib()
    block: Block = attr.ib()



from __future__ import annotations

import copy
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Union

from terra_proto.cosmos.base.v1beta1 import Coin as Coin_pb

from terra_sdk.util.json import JSONSerializable

from .coin import Coin
from .numeric import Numeric


class Coins(JSONSerializable, List[Coin_pb]):
    """Represents an unordered collection of :class:`Coin` objects
    -- analagous to ``sdk.Coins`` and ``sdk.DecCoins`` in Cosmos SDK. If one of the
    input coins would be ``Dec``-amount type coin, the resultant Coins is converted to
    ``Dec``-amount coins.

    Args:
        arg (Optional[Coins.Input], optional): argument to convert. Defaults to ``{}``.

    Raises:
        TypeError: if ``arg`` is not an Iterable
    """

    Input = Union[Iterable[Coin], str, Dict[str, Numeric.Input], Dict[str, Coin]]
    """Types which can be converted into a :class:`Coins` object."""

    _coins: Dict[str, Coin]

    def __repr__(self) -> str:
        if len(self) == 0:
            return "Coins()"
        else:
            return f"Coins('{self!s}')"

    def __str__(self) -> str:
        return ",".join(str(coin) for coin in self)

    @classmethod
    def from_str(cls, s: str) -> Coins:
        """Converts a comma-separated list of Coin-format strings to :class:`Coins`.

        >>> Coins.from_str('1000uluna,1234ukrw')
        Coins("1000uluna,1234ukrw")

        Args:
            s (str): string to convert
        """
        coin_strings = s.split(r",")
        return Coins(Coin.from_str(cs) for cs in coin_strings)

    def __init__(self, arg: Optional[Coins.Input] = {}, **denoms):
        """Converts the argument into a :class:`Coins` object."""

        if arg is None:
            self._coins = {}
            return

        # arg should be an iterable
        try:
            iter(arg)
        except TypeError:
            raise TypeError(f"could not create Coins object with argument: {arg!s}")

        if isinstance(arg, Coins):
            self._coins = copy.deepcopy(arg._coins)
            return

        if isinstance(arg, str):
            self._coins = Coins.from_str(arg)._coins
            return

        self._coins = Coins(denoms)._coins if denoms else {}

        coins: Iterable[Coin]
        if isinstance(arg, dict):
            coins = [Coin(denom, arg[denom]) for denom in arg]
        else:
            coins = arg
        for coin in coins:
            x = self._coins.get(coin.denom)
            if x is not None:
                self._coins[coin.denom] = x + coin
            else:
                self._coins[coin.denom] = coin

        # make all coins DecCoin if one is DecCoin
        if not all([c.is_int_coin() for c in self]):
            for denom in self._coins:
                self._coins[denom] = self._coins[denom].to_dec_coin()

    def __getitem__(self, denom: str) -> Coin:
        return self._coins[denom]

    def get(self, denom: str) -> Optional[Coin]:
        """Get the Coin with the denom contained in the Coins set.

        Args:
            denom (str): denom

        Returns:
            Optional[Coin]: result (can be ``None``)
        """
        return self._coins.get(denom)

    @classmethod
    def from_data(cls, data: list) -> Coins:
        """Converts list of Coin-data objects to :class:`Coins`.

        Args:
            data (list): list of Coin-data objects
        """
        coins = map(Coin.from_data, data)
        return cls(coins)

    @classmethod
    def from_amino(cls, amino: list) -> Coins:
        """Converts list of Coin-amino objects to :class:`Coins`.

        Args:
            amino (list): list of Coin-data objects
        """
        coins = map(Coin.from_amino, amino)
        return cls(coins)

    def to_amino(self) -> List[dict]:
        return [coin.to_amino() for coin in self]

    def to_data(self) -> List[dict]:
        return [coin.to_data() for coin in self]

    @classmethod
    def from_proto(cls, proto: List[Coin_pb]) -> Coins:
        """Converts list of Coin-data objects to :class:`Coins`.

        Args:
            data (list): list of Coin-data objects
        """
        coins = map(Coin.from_proto, proto)
        return cls(coins)

    def to_proto(self) -> List[Coin_pb]:
        return [coin.to_proto() for coin in self]

    def to_dict(self) -> List[dict]:
        return [coin.to_dict for coin in self]

    def denoms(self) -> List[str]:
        """Get the list of denoms for all Coin objects contained."""
        return [c.denom for c in self]

    def to_dec_coins(self) -> Coins:
        """Creates new set of :class:`Coins` that have :class`Dec` amounts."""
        return Coins(c.to_dec_coin() for c in self)

    def to_int_coins(self) -> Coins:
        """Creates new set of :class:`Coins` that have ``int`` amounts."""
        return Coins(c.to_int_coin() for c in self)

    def to_int_ceil_coins(self) -> Coins:
        """Creates a new :class:`Coins` object with all ``int`` coins with ceiling the amount"""
        return Coins(c.to_int_ceil_coin() for c in self)

    def add(self, addend: Union[Coin, Coins]) -> Coins:
        """Performs addition, which combines the sets of Coin objects. Coins of similar denoms
        will be merged into one Coin representing the denom.

        Args:
            addend (Union[Coin, Coins]): addend
        """
        if isinstance(addend, Coin):
            return Coins([addend, *self.to_list()])
        else:
            return Coins([*addend.to_list(), *self.to_list()])

    def __add__(self, addend: Union[Coin, Coins]) -> Coins:
        return self.add(addend)

    def sub(self, subtrahend: Union[Coin, Coins]) -> Coins:
        """Performs subtraction, which combines the sets of Coin objects. Coins of similar denoms
        will be merged into one Coin representing the denom.

        Args:
            subtrahend (Union[Coin, Coins]): subtrahend
        """
        return self.add(subtrahend.mul(-1))

    def __sub__(self, subtrahend: Union[Coin, Coins]) -> Coins:
        return self.sub(subtrahend)

    def mul(self, multiplier: Numeric.Input) -> Coins:
        """Performs multiplicaiton, which multiplies all the Coin objects in the set by a
        multiplier.

        Args:
            multiplier (Numeric.Input): multiplier
        """
        return Coins(coin.mul(multiplier) for coin in self)

    def __mul__(self, multiplier: Numeric.Input) -> Coins:
        return self.mul(multiplier)

    def div(self, divisor: Numeric.Input) -> Coins:
        """Performs division, which divides all the Coin objects in the set by a divisor.

        Args:
            divisor (Numeric.Input): divisor
        """
        return Coins(coin.div(divisor) for coin in self)

    def __truediv__(self, divisor: Numeric.Input) -> Coins:
        return Coins(coin / divisor for coin in self)

    def __floordiv__(self, divisor: Numeric.Input) -> Coins:
        return Coins(coin // divisor for coin in self)

    def to_list(self) -> List[Coin]:
        """Converts the set of :class:`Coin` objects contained into a sorted list by denom.

        Returns:
            List[Coin]: list, sorted by denom
        """
        return sorted(self._coins.values(), key=lambda c: c.denom)

    def filter(self, predicate: Callable[[Coin], bool]) -> Coins:
        """Creates a new :class:`Coins` collection which filters out all Coin objects that
        do not meet the predicate.

        Args:
            predicate (Callable[[Coin], bool]): predicate for filtering
        """
        return Coins(c for c in self if predicate(c))

    def map(self, fn: Callable[[Coin], Any]) -> Iterator[Any]:
        """Creates an iterable which applies the function to all coins in the set,
        ordered by denomination.

        Args:
            fn (Callable[[Coin], Any]): function to apply

        Returns:
            Iterator[Any]: coin map

        Yields:
            Iterator[Any]: coin map
        """
        return map(fn, self)

    def __eq__(self, other) -> bool:
        try:
            return self.to_list() == other.to_list()
        except AttributeError:
            return False

    def __iter__(self):
        return iter(self.to_list())

    def __len__(self):
        return len(self.to_list())

    def __contains__(self, denom: str) -> bool:
        return denom in self._coins



from __future__ import annotations

import base64
import hashlib
from abc import ABC, abstractmethod
from typing import List, Union

import attr
from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.cosmos.crypto.ed25519 import PubKey as ValConsPubKey_pb
from terra_proto.cosmos.crypto.multisig import LegacyAminoPubKey as LegacyAminoPubKey_pb
from terra_proto.cosmos.crypto.secp256k1 import PubKey as SimplePubKey_pb

from terra_sdk.util.json import JSONSerializable

from ..util.base import create_demux_unpack_any
from .bech32 import get_bech

BECH32_AMINO_PUBKEY_DATA_PREFIX_SECP256K1 = "eb5ae987" + "21"  # with fixed length 21
BECH32_AMINO_PUBKEY_DATA_PREFIX_ED25519 = "1624de64" + "20"  # with fixed length 20
BECH32_AMINO_PUBKEY_DATA_PREFIX_MULTISIG_THRESHOLD = "22c1f7e2"  # without length


__all__ = [
    "PublicKey",
    "SimplePublicKey",
    "ValConsPubKey",
    "LegacyAminoMultisigPublicKey",
    "address_from_public_key",
    "amino_pubkey_from_public_key",
]


def encode_uvarint(value: Union[int, str]) -> List[int]:
    val = int(str(value))
    if val > 127:
        raise ValueError(
            "Encoding numbers > 127 is not supported here. Please tell those lazy CosmJS maintainers to "
            "port the binary.PutUvarint implementation from the Go standard library and write some "
            "tests."
        )
    return [val]


def address_from_public_key(public_key: PublicKey) -> bytes:
    sha = hashlib.sha256()
    rip = hashlib.new("ripemd160")
    sha.update(public_key.key)
    rip.update(sha.digest())
    return rip.digest()


def amino_pubkey_from_public_key(public_key: PublicKey) -> bytes:
    arr = bytes.fromhex(BECH32_AMINO_PUBKEY_DATA_PREFIX_SECP256K1)
    arr += bytes(public_key.key)
    return bytes(arr)


class PublicKey(JSONSerializable, ABC):
    """Data object holding the public key component of an account or signature."""

    @property
    @abstractmethod
    def type_url(self):
        pass

    @property
    @abstractmethod
    def type_amino(self):
        pass

    @abstractmethod
    def get_type(self) -> str:
        return self.type_url

    @classmethod
    def from_proto(cls, proto: Any_pb):
        type_url = proto.type_url
        value = proto.value
        if type_url == SimplePublicKey.type_url:
            return SimplePublicKey.from_proto(SimplePubKey_pb().parse(value))
        elif type_url == ValConsPubKey.type_url:
            return ValConsPubKey.from_proto(ValConsPubKey_pb().parse(value))
        elif type_url == LegacyAminoMultisigPublicKey.type_url:
            return LegacyAminoMultisigPublicKey.from_proto(
                LegacyAminoPubKey_pb().parse(value)
            )
        raise TypeError("could not marshal PublicKey: type is incorrect")

    @classmethod
    def from_amino(cls, amino: dict):
        type_amino = amino.get("type")
        if type_amino == SimplePublicKey.type_amino:
            return SimplePublicKey.from_amino(amino)
        elif type_amino == ValConsPubKey.type_amino:
            return ValConsPubKey.from_amino(amino)
        elif type_amino == LegacyAminoMultisigPublicKey.type_amino:
            return LegacyAminoMultisigPublicKey.from_amino(amino)
        raise TypeError("could not marshal PublicKey: type is incorrect")

    @classmethod
    def from_data(cls, data: dict):
        type_url = data["@type"]
        if type_url == SimplePublicKey.type_url:
            return SimplePublicKey.from_data(data)
        elif type_url == ValConsPubKey.type_url:
            return ValConsPubKey.from_data(data)
        elif type_url == LegacyAminoMultisigPublicKey.type_url:
            return LegacyAminoMultisigPublicKey.from_data(data)
        raise TypeError("could not unmarshal PublicKey: type is incorrect")

    @classmethod
    def unpack_any(cls, any_pb: Any_pb):
        unpack = create_demux_unpack_any(
            [SimplePublicKey, ValConsPubKey, LegacyAminoMultisigPublicKey]
        )
        return unpack(any_pb)

    @abstractmethod
    def pack_any(self) -> Any_pb:
        raise NotImplementedError

    @abstractmethod
    def address(self) -> str:
        pass

    @abstractmethod
    def raw_address(self) -> str:
        pass

    @abstractmethod
    def encode_amino_pubkey(self) -> bytes:
        pass

    @abstractmethod
    def to_amino(self) -> dict:
        pass

    @abstractmethod
    def to_data(self) -> dict:
        pass

    @abstractmethod
    def to_proto(self):
        pass


@attr.s
class SimplePublicKey(PublicKey):
    """Data object holding the SIMPLE public key component of an account or signature."""

    type_amino = "tendermint/PubKeySecp256k1"
    """"""
    type_url = "/cosmos.crypto.secp256k1.PubKey"
    """Normal signature public key type."""
    prototype = SimplePubKey_pb
    """"""

    key: bytes = attr.ib()

    def to_amino(self) -> dict:
        return {"type": self.type_amino, "value": base64.b64encode(self.key)}

    def to_data(self) -> dict:
        return {"@type": self.type_url, "key": base64.b64encode(self.key)}

    @classmethod
    def from_data(cls, data: dict) -> SimplePublicKey:
        return cls(key=base64.b64decode(data["key"]))

    @classmethod
    def from_proto(cls, proto: SimplePubKey_pb) -> SimplePublicKey:
        return cls(key=proto.key)

    @classmethod
    def from_amino(cls, amino: dict) -> SimplePublicKey:
        return cls(key=base64.b64decode(amino["value"]))

    def to_proto(self) -> SimplePubKey_pb:
        return SimplePubKey_pb(key=self.key)

    def get_type(self) -> str:
        return self.type_url

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    def encode_amino_pubkey(self) -> bytearray:
        out = bytearray.fromhex(BECH32_AMINO_PUBKEY_DATA_PREFIX_SECP256K1) + bytearray(
            self.key
        )
        return out

    def raw_address(self) -> bytes:
        return address_from_public_key(self)

    def address(self) -> str:
        return get_bech("terra", self.raw_address().hex())


@attr.s
class ValConsPubKey(PublicKey):
    """Data object holding the public key component of a validator's account or signature."""

    type_amino = "tendermint/PubKeyEd25519"
    """"""
    type_url = "/cosmos.crypto.ed25519.PubKey"
    """an ed25519 tendermint public key type."""
    prototype = ValConsPubKey_pb
    """"""

    key: bytes = attr.ib()

    def to_amino(self) -> dict:
        return {"type": self.type_amino, "value": base64.b64encode(self.key)}

    def to_data(self) -> dict:
        return {"@type": self.type_url, "key": base64.b64encode(self.key)}

    @classmethod
    def from_data(cls, data: dict) -> ValConsPubKey:
        return cls(key=base64.b64decode(data["key"]))

    @classmethod
    def from_amino(cls, amino: dict) -> ValConsPubKey:
        return cls(key=base64.b64decode(amino["value"]))

    @classmethod
    def from_proto(cls, proto: ValConsPubKey_pb) -> ValConsPubKey:
        return cls(key=proto.key)

    def get_type(self) -> str:
        return self.type_url

    def to_proto(self) -> ValConsPubKey_pb:
        return ValConsPubKey_pb(key=base64.b64encode(self.key))

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    def encode_amino_pubkey(self) -> bytes:
        return bytes.fromhex(BECH32_AMINO_PUBKEY_DATA_PREFIX_ED25519) + bytes(self.key)

    def raw_address(self) -> bytes:
        return address_from_public_key(self)

    def address(self) -> str:
        return get_bech("terravalcons", self.raw_address().hex())


@attr.s
class LegacyAminoMultisigPublicKey(PublicKey):
    """Data object holding the Legacy Amino-typed public key component of an account or signature."""

    type_amino = "tendermint/PubKeyMultisigThreshold"
    """"""
    type_url = "/cosmos.crypto.multisig.LegacyAminoPubKey"
    """Multisig public key type."""
    prototype = LegacyAminoPubKey_pb
    """"""

    threshold: int = attr.ib(converter=int)
    public_keys: List[SimplePublicKey] = attr.ib(factory=list)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "threshold": str(self.threshold),
                "pubkeys": [pubkey.to_amino() for pubkey in self.public_keys],
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "threshold": self.threshold,
            "public_keys": self.public_keys,
        }

    @classmethod
    def from_data(cls, data: dict) -> LegacyAminoMultisigPublicKey:
        return cls(
            threshold=data["threshold"],
            public_keys=[PublicKey.from_data(pubkey) for pubkey in data["public_keys"]],
        )

    @classmethod
    def from_proto(cls, proto: LegacyAminoPubKey_pb) -> LegacyAminoMultisigPublicKey:
        return cls(
            threshold=proto.threshold,
            public_keys=[PublicKey.unpack_any(pk) for pk in proto.public_keys],
        )

    @classmethod
    def from_amino(cls, amino: dict) -> LegacyAminoMultisigPublicKey:
        return cls(
            threshold=amino["value"]["threshold"],
            public_keys=[
                SimplePublicKey.from_amino(pubkey)
                for pubkey in amino["value"]["public_keys"]
            ],
        )

    def get_type(self) -> str:
        return self.type_url

    def to_proto(self) -> LegacyAminoPubKey_pb:
        return LegacyAminoPubKey_pb(
            threshold=self.threshold,
            public_keys=[pk.pack_any() for pk in self.public_keys],
        )

    def encode_amino_pubkey(self) -> bytearray:
        if self.threshold > 127:
            raise ValueError("threshold over 127 is now supported here")
        out = bytearray.fromhex(BECH32_AMINO_PUBKEY_DATA_PREFIX_MULTISIG_THRESHOLD)
        out.append(0x08)
        out += bytearray(encode_uvarint(self.threshold))
        for pkData in [pubkey.encode_amino_pubkey() for pubkey in self.public_keys]:
            out.append(0x12)
            out += bytearray(encode_uvarint(len(pkData)))
            out += pkData
        return out

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    def raw_address(self) -> str:
        pubkey_data = bytes(self.encode_amino_pubkey())
        hasher = hashlib.sha256()
        hasher.update(pubkey_data)
        return hasher.digest()[0:20].hex()

    def address(self) -> str:
        address = get_bech("terra", self.raw_address())
        return address

    def pubkey_address(self) -> str:
        return get_bech("terrapub", str(self.encode_amino_pubkey()))



"""Useful contract-related functions."""

import base64
from typing import Dict, List, Union

from terra_sdk.core import AccAddress
from terra_sdk.core.auth import TxInfo
from terra_sdk.core.broadcast import BlockTxBroadcastResult

__all__ = [
    "read_file_as_b64",
    "get_code_id",
    "get_contract_address",
    "get_contract_events",
]


def read_file_as_b64(path: Union[str, bytes, int]) -> str:
    """Reads a file's contents as binary bytes and encodes it in a base64-string.

    Args:
        path (Union[str, bytes, int]): binary file path

    Returns:
        str: file's bytes in base64-encoded string
    """
    with open(path, "rb") as contract_file:
        contract_bytes = base64.b64encode(contract_file.read()).decode()
        return contract_bytes


def get_code_id(
    tx_result: Union[BlockTxBroadcastResult, TxInfo], msg_index: int = 0
) -> str:
    """Utility function for extracting the code id from a ``MsgStoreCode`` message.

    Args:
        tx_result (BlockTxBroadcastResult): broadcast result
        msg_index (int, optional): index of ``MsgStoreCode`` inside tx. Defaults to 0.

    Returns:
        str: extracted code id
    """
    if tx_result.logs:
        code_id = tx_result.logs[msg_index].events_by_type["store_code"]["code_id"][0]
        return code_id
    else:
        raise ValueError("could not parse code id -- tx logs are empty.")


def get_contract_address(
    tx_result: Union[BlockTxBroadcastResult, TxInfo], msg_index: int = 0
) -> AccAddress:
    """Utility function for extracting the contract address from a ``MsgInstantiateContract``
    message.

    Args:
        tx_result (BlockTxBroadcastResult): broadcast result
        msg_index (int, optional): index of ``MsgInstantiateContract`` inside tx. Defaults to 0.

    Returns:
        str: extracted contract address
    """
    if tx_result.logs:
        contract_address = tx_result.logs[msg_index].events_by_type["wasm"][
            "_contract_address"
        ][0]
        return AccAddress(contract_address)
    else:
        raise ValueError("could not parse code id -- tx logs are empty.")


def get_contract_events(
    tx_result: Union[BlockTxBroadcastResult, TxInfo], msg_index: int = 0
) -> List[Dict[str, str]]:
    if tx_result.logs:
        contract_events = []
        for event in tx_result.logs[msg_index].events:
            if event["type"] == "from_contract":
                event_data: Dict[str, str] = {}
                current_contract_address = event["attributes"][0]["value"]
                for att in event["attributes"]:
                    if (
                        att["key"] == "contract_address"
                        and current_contract_address != att["value"]
                    ):
                        contract_events.append(event_data)
                        event_data = {}
                        current_contract_address = att["value"]
                    event_data[att["key"]] = att["value"]
                contract_events.append(event_data)  # append remaining
                return contract_events
        raise ValueError("could not find event type 'from_contract' in logs")
    else:
        raise ValueError("could not parse contract events -- tx logs are empty.")



"""Data objects about Signature V2."""

from __future__ import annotations

from typing import List, Optional, Tuple

import attr
from terra_proto.cosmos.crypto.multisig.v1beta1 import (
    MultiSignature as MultiSignature_pb,
)
from terra_proto.cosmos.tx.signing.v1beta1 import SignMode

from .compact_bit_array import CompactBitArray
from .mode_info import ModeInfo, ModeInfoMulti, ModeInfoSingle
from .public_key import PublicKey

__all__ = ["SignatureV2", "Descriptor", "Single", "Multi", "SignMode"]


@attr.s
class SignatureV2:
    public_key: PublicKey = attr.ib()
    data: Descriptor = attr.ib()
    sequence: int = attr.ib(converter=int)

    @classmethod
    def from_data(cls, data: dict) -> SignatureV2:
        return cls(
            public_key=PublicKey.from_data(data["public_key"]),
            data=Descriptor.from_data(data["data"]),
            sequence=data["sequence"],
        )

    def to_data(self) -> dict:
        return {
            "public_key": self.public_key.to_data(),
            "data": self.data.to_data(),
            "sequence": self.sequence,
        }


@attr.s
class Descriptor:
    single: Optional[Single] = attr.ib(default=None)
    multi: Optional[Multi] = attr.ib(default=None)

    @classmethod
    def from_data(cls, data: dict) -> Descriptor:
        s = None
        m = None
        if data["single"] is not None:
            s = Single.from_data(data["single"])
        if data["multi"] is not None:
            m = Multi.from_data(data["multi"])
        return cls(single=s, multi=m)

    def to_data(self) -> dict:
        typ = "single" if self.single else "multi"
        dat = self.single.to_data() if self.single else self.multi.to_data()
        return {typ: dat}

    def to_mode_info_and_signature(self) -> Tuple[ModeInfo, bytes]:
        if self.single is not None:
            sig_data = self.single
            return [ModeInfo(single=ModeInfoSingle(sig_data.mode)), sig_data.signature]

        if self.multi:
            sig_data = self.multi
            mode_infos: List[ModeInfo] = []
            signatures: List[bytes] = []
            for sig in sig_data.signatures:
                mode_info, sig_bytes = sig.to_mode_info_and_signature()
                mode_infos.append(mode_info)
                signatures.append(sig_bytes)
            pb = MultiSignature_pb(signatures=signatures)
            return [
                ModeInfo(
                    multi=ModeInfoMulti(
                        bitarray=sig_data.bitarray, mode_infos=mode_infos
                    )
                ),
                bytes(pb),  # base64.b64encode(bytes(pb)),
            ]

        raise ValueError("invalid signature descriptor")


@attr.s
class Single:  # FIXME: SignModeTo/FromJSON
    mode: SignMode = attr.ib()
    signature: bytes = attr.ib()

    @classmethod
    def from_data(cls, data: dict) -> Single:
        return cls(mode=data["mode"], signature=data["signature"])

    def to_data(self) -> dict:
        return {"mode": self.mode, "signature": self.signature}


@attr.s
class Multi:
    bitarray: CompactBitArray = attr.ib()
    signatures: List[Descriptor] = attr.ib()

    @classmethod
    def from_data(cls, data: dict) -> Multi:
        return cls(
            CompactBitArray.from_data(data["bitarray"]),
            [Descriptor.from_data(d) for d in data["signatures"]],
        )

    def to_data(self) -> dict:
        return {
            "bitarray": self.bitarray.to_data(),
            "signatures": [sig.to_data() for sig in self.signatures],
        }



"""Distribution module governance proposal types."""

from __future__ import annotations

import attr
from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.cosmos.distribution.v1beta1 import (
    CommunityPoolSpendProposal as CommunityPoolSpendProposal_pb,
)

from terra_sdk.core import AccAddress, Coins
from terra_sdk.util.json import JSONSerializable

__all__ = ["CommunityPoolSpendProposal"]


@attr.s
class CommunityPoolSpendProposal(JSONSerializable):
    """Proposal for allocating funds from the community pool to an address.

    Args:
        title: proposal title
        description: proposal description
        recipient: designated recipient of funds if proposal passes
        amount (Coins): amount to spend from community pool
    """

    type_amino = "cosmos-sdk/CommunityPoolSpendProposal"
    """"""
    type_url = "/cosmos.distribution.v1beta1.CommunityPoolSpendProposal"
    """"""
    prototype = CommunityPoolSpendProposal_pb
    """"""

    title: str = attr.ib()
    description: str = attr.ib()
    recipient: AccAddress = attr.ib()
    amount: Coins = attr.ib(converter=Coins)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "title": self.title,
                "description": self.description,
                "recipient": self.recipient,
                "amount": self.amount.to_amino(),
            },
        }

    @classmethod
    def from_data(cls, data: dict) -> CommunityPoolSpendProposal:
        return cls(
            title=data["title"],
            description=data["description"],
            recipient=data["recipient"],
            amount=Coins.from_data(data["amount"]),
        )

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "title": self.title,
            "description": self.description,
            "recipient": self.recipient,
            "amount": self.amount.to_data(),
        }

    def to_proto(self) -> CommunityPoolSpendProposal_pb:
        return CommunityPoolSpendProposal_pb(
            title=self.title,
            description=self.description,
            recipient=self.recipient,
            amount=self.amount.to_proto(),
        )

    @classmethod
    def from_proto(
        cls, proto: CommunityPoolSpendProposal_pb
    ) -> CommunityPoolSpendProposal:
        return cls(
            title=proto.title,
            description=proto.description,
            recipient=proto.recipient,
            amount=Coins.from_proto(proto.amount),
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))



"""Data objects about Fee."""

from __future__ import annotations

from typing import Optional

import attr
from terra_proto.cosmos.tx.v1beta1 import Fee as Fee_pb

from terra_sdk.core.bech32 import AccAddress
from terra_sdk.core.coins import Coins
from terra_sdk.util.json import JSONSerializable

__all__ = ["Fee"]


@attr.s
class Fee(JSONSerializable):
    """Data structure holding information for a transaction fee.

    Args:
        gas (int): gas to use ("gas requested")
        amount (Coins.Input): fee amount
        payer (AccAddress, optional): address of fee payer
        granter (AccAddress, optional): address of fee granter
    """

    gas_limit: int = attr.ib(converter=int)
    amount: Coins = attr.ib(converter=Coins)
    payer: Optional[AccAddress] = attr.ib(default=None)
    granter: Optional[AccAddress] = attr.ib(default=None)

    def to_amino(self) -> dict:
        return {"gas": str(self.gas_limit), "amount": self.amount.to_amino()}

    @classmethod
    def from_data(cls, data: dict) -> Fee:
        return cls(
            int(data["gas_limit"]),
            Coins.from_data(data["amount"]),
            data["payer"],
            data["granter"],
        )

    def to_data(self) -> dict:
        return {
            "gas_limit": str(self.gas_limit),
            "amount": self.amount.to_data(),
            "payer": str(self.payer),
            "granter": str(self.granter),
        }

    def to_proto(self) -> Fee_pb:
        return Fee_pb(
            amount=self.amount.to_proto(),
            gas_limit=self.gas_limit,
            payer=self.payer,
            granter=self.granter,
        )

    @classmethod
    def from_proto(cls, proto: Fee_pb) -> Fee:
        return cls(
            gas_limit=proto.gas_limit,
            amount=Coins.from_proto(proto.amount),
            payer=proto.payer,
            granter=proto.granter,
        )

    @property
    def gas_prices(self) -> Coins:
        return self.amount.to_dec_coins().div(self.gas_limit)



from __future__ import annotations

import hashlib

from ecdsa import SECP256k1, SigningKey
from ecdsa.util import sigencode_string_canonize

from .key import Key

__all__ = ["RawKey"]

from ..core import PublicKey, SimplePublicKey


def compute_public_key(private_key: bytes) -> PublicKey:
    return SimplePublicKey(
        key=SigningKey.from_string(private_key, curve=SECP256k1)
        .get_verifying_key()
        .to_string("compressed")
    )


class RawKey(Key):
    """RawKey directly uses a raw (plaintext) private key in memory, and provides
    the implementation for signing with ECDSA on curve Secp256k1.

    Args:
        private_key (bytes): private key in bytes
    """

    private_key: bytes
    """Private key, in bytes."""

    @classmethod
    def from_hex(cls, private_key_hex: str) -> RawKey:
        """Create a new RawKey from a hex-encoded private key string.

        Args:
            private_key_hex (str): hex-encoded private key
        """
        return cls(bytes.fromhex(private_key_hex))

    def __init__(self, private_key: bytes):
        public_key = compute_public_key(private_key)
        super().__init__(public_key)
        self.private_key = private_key

    def sign(self, payload: bytes) -> bytes:
        """Signs the data payload using ECDSA and curve Secp256k1 with the private key as
        the signing key.

        Args:
            payload (bytes): data to sign
        """
        sk = SigningKey.from_string(self.private_key, curve=SECP256k1)
        return sk.sign_deterministic(
            payload,
            hashfunc=hashlib.sha256,
            sigencode=sigencode_string_canonize,
        )



from typing import Union

from boltons.iterutils import remap  # type: ignore

__all__ = ["remove_none"]


def remove_none(obj: Union[dict, str]):
    """remove keys for None in a dict"""
    return remap(
        obj, visit=lambda path, key, value: key is not None and value is not None
    )



"""ibc connection module message types."""

from __future__ import annotations

from typing import List

import attr
from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.ibc.core.connection.v1 import (
    MsgConnectionOpenAck as MsgConnectionOpenAck_pb,
)
from terra_proto.ibc.core.connection.v1 import (
    MsgConnectionOpenConfirm as MsgConnectionOpenConfirm_pb,
)
from terra_proto.ibc.core.connection.v1 import (
    MsgConnectionOpenInit as MsgConnectionOpenInit_pb,
)
from terra_proto.ibc.core.connection.v1 import (
    MsgConnectionOpenTry as MsgConnectionOpenTry_pb,
)

from terra_sdk.core import AccAddress
from terra_sdk.core.ibc.data import Height
from terra_sdk.core.ibc.data.connection import Counterparty, Version
from terra_sdk.core.msg import Msg

__all__ = [
    "MsgConnectionOpenInit",
    "MsgConnectionOpenTry",
    "MsgConnectionOpenAck",
    "MsgConnectionOpenConfirm",
]


@attr.s
class MsgConnectionOpenInit(Msg):
    """
    MsgConnectionOpenInit defines the msg sent by an account on Chain A to initialize a connection with Chain B.
    """

    type_url = "/ibc.core.connection.v1.MsgConnectionOpenInit"
    """"""
    prototype = MsgConnectionOpenInit_pb
    """"""

    client_id: str = attr.ib()
    counterparty: Counterparty = attr.ib()
    version: Version = attr.ib()
    delay_period: int = attr.ib(converter=int)
    signer: AccAddress = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgConnectionOpenInit:
        return cls(
            client_id=data["client_id"],
            counterparty=Counterparty.from_data(data["counterparty"]),
            version=Version.from_data(data["version"]),
            delay_period=data["delay_period"],
            signer=data["signer"],
        )

    def to_proto(self) -> MsgConnectionOpenInit_pb:
        return MsgConnectionOpenInit_pb(
            client_id=self.client_id,
            counterparty=self.counterparty.to_proto(),
            version=self.version.to_proto(),
            delay_period=self.delay_period,
            signer=self.signer,
        )

    @classmethod
    def from_proto(cls, proto: MsgConnectionOpenInit_pb) -> MsgConnectionOpenInit:
        return cls(
            client_id=proto.client_id,
            counterparty=Counterparty.from_proto(proto.counterparty),
            version=Version.from_proto(proto.version),
            delay_period=proto.delay_period,
            signer=proto.signer,
        )


@attr.s
class MsgConnectionOpenTry(Msg):
    """
    MsgConnectionOpenTry defines a msg sent by a Relayer to try to open a connection on Chain B.
    """

    type_url = "/ibc.core.connection.v1.MsgConnectionOpenTry"
    """"""
    prototype = MsgConnectionOpenTry_pb
    """"""

    client_id: str = attr.ib()
    previous_connection_id: str = attr.ib()
    client_state: dict = attr.ib()
    counterparty: Counterparty = attr.ib()
    delay_period: int = attr.ib(converter=int)
    counterparty_versions: List[Version] = attr.ib(converter=list)
    proof_height: Height = attr.ib()
    proof_init: bytes = attr.ib()
    proof_client: bytes = attr.ib()
    proof_consensus: bytes = attr.ib()
    consensus_height: Height = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgConnectionOpenTry:
        return cls(
            client_id=data["client_id"],
            previous_connection_id=data["previous_connection_id"],
            client_state=data["client_state"],
            counterparty=Counterparty.from_data(data["counterparty"]),
            delay_period=data["delay_period"],
            counterparty_versions=[
                Version.from_data(ver) for ver in data["counterparty_versions"]
            ],
            proof_height=Height.from_data(data["proof_height"]),
            proof_init=data["proof_init"],
            proof_client=data["proof_client"],
            proof_consensus=data["proof_consensus"],
            consensus_height=Height.from_data(data["consensus_height"]),
            signer=data["signer"],
        )

    def to_proto(self) -> MsgConnectionOpenTry_pb:
        return MsgConnectionOpenTry_pb(
            client_id=self.client_id,
            previous_connection_id=self.previous_connection_id,
            client_state=Any_pb().from_dict(self.client_state),
            counterparty=self.counterparty.to_proto(),
            delay_period=self.delay_period,
            counterparty_versions=[
                ver.to_proto() for ver in self.counterparty_versions
            ],
            proof_height=self.proof_height.to_proto(),
            proof_init=self.proof_init,
            proof_client=self.proof_client,
            proof_consensus=self.proof_consensus,
            consensus_height=self.consensus_height.to_proto(),
            signer=self.signer,
        )

    @classmethod
    def from_proto(cls, proto: MsgConnectionOpenTry_pb) -> MsgConnectionOpenTry:
        return cls(
            client_id=proto.client_id,
            previous_connection_id=proto.previous_connection_id,
            client_state=proto.client_state,
            counterparty=Counterparty.from_proto(proto.counterparty),
            delay_period=proto.delay_period,
            counterparty_versions=[
                Version.from_proto(ver) for ver in proto.counterparty_versions
            ],
            proof_height=Height.from_proto(proto.proof_height),
            proof_init=proto.proof_init,
            proof_client=proto.proof_client,
            proof_consensus=proto.proof_consensus,
            consensus_height=Height.from_proto(proto.consensus_height),
            signer=proto.signer,
        )


@attr.s
class MsgConnectionOpenAck(Msg):
    """
    MsgConnectionOpenAck defines a msg sent by a Relayer to Chain A to acknowledge the change of connection state to TRYOPEN on Chain B.
    """

    type_url = "/ibc.core.connection.v1.MsgConnectionOpenAck"
    """"""
    prototype = MsgConnectionOpenAck_pb
    """"""

    connection_id: str = attr.ib()
    counterparty_connection_id: str = attr.ib()
    version: Version = attr.ib()
    client_state: any = attr.ib()
    proof_height: Height = attr.ib()
    proof_try: bytes = attr.ib()
    proof_client: bytes = attr.ib()
    proof_consensus: bytes = attr.ib()
    consensus_height: Height = attr.ib()
    signer: str = attr.ib()

    def to_amino(self):
        raise Exception("Amino not supported")

    @classmethod
    def from_data(cls, data: dict) -> MsgConnectionOpenAck:
        return cls(
            connection_id=data["connection_id"],
            counterparty_connection_id=data["counterparty_connection_id"],
            version=Version.from_data(data["version"]),
            client_state=data["client_state"],
            proof_height=Height.from_data(data["proof_height"]),
            proof_try=data["proof_try"],
            proof_client=data["proof_client"],
            proof_consensus=data["proof_consensus"],
            consensus_height=Height.from_data(data["consensus_height"]),
            signer=data["signer"],
        )

    def to_proto(self) -> MsgConnectionOpenAck_pb:
        return MsgConnectionOpenAck_pb(
            connection_id=self.connection_id,
            counterparty_connection_id=self.counterparty_connection_id,
            version=self.version.to_proto(),
            client_state=Any_pb().from_dict(self.client_state),
            proof_height=self.proof_height.to_proto(),
            proof_try=self.proof_try,
            proof_client=self.proof_client,
            proof_consensus=self.proof_consensus,
            consensus_height=self.consensus_height.to_proto(),
            signer=self.signer,
        )

    @classmethod
    def from_proto(cls, proto: MsgConnectionOpenAck_pb) -> MsgConnectionOpenAck:
        return cls(
            connection_id=proto.connection_id,
            counterparty_connection_id=proto.counterparty_connection_id,
            version=Version.from_proto(proto.version),
            client_state=Any_pb().parse(proto.client_state),
            proof_height=Height.from_proto(proto.proof_height),
            proof_try=proto.proof_try,
            proof_client=proto.proof_client,
            proof_consensus=proto.proof_consensus,
            consensus_height=Height.from_proto(proto.consensus_height),
            signer=proto.signer,
        )


@attr.s
class MsgConnectionOpenConfirm(Msg):
    """
    MsgConnectionOpenConfirm defines a msg sent by a Relayer to Chain B to acknowledge the change of connection state to OPEN on Chain A.
    """

    type_url = "/ibc.core.connection.v1.MsgConnectionOpenConfirm"
    """"""
    prototype = MsgConnectionOpenConfirm_pb
    """"""

    def to_amino(self):
        raise Exception("Amino not supported")

    connection_id: str = attr.ib()
    proof_ack: bytes = attr.ib()
    proof_height: Height = attr.ib()
    signer: str = attr.ib()

    @classmethod
    def from_data(cls, data: dict) -> MsgConnectionOpenConfirm:
        return cls(
            connection_id=data["connection_id"],
            proof_ack=data["proof_ack"],
            proof_height=Height.from_data(data["proof_height"]),
            signer=data["signer"],
        )

    def to_proto(self) -> MsgConnectionOpenConfirm_pb:
        return MsgConnectionOpenConfirm_pb(
            connection_id=self.connection_id,
            proof_ack=self.proof_ack,
            proof_height=self.proof_height.to_proto(),
            signer=self.signer,
        )

    @classmethod
    def from_proto(cls, proto: MsgConnectionOpenConfirm_pb) -> MsgConnectionOpenConfirm:
        return cls(
            connection_id=proto.connection_id,
            proof_ack=proto.proof_ack,
            proof_height=Height.from_proto(proto.proof_height),
            signer=proto.signer,
        )



from typing import Union

from terra_sdk.core import AccAddress
from terra_sdk.core.auth import (
    Account,
    BaseAccount,
    ContinuousVestingAccount,
    DelayedVestingAccount,
    PeriodicVestingAccount,
)

from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncAuthAPI", "AuthAPI"]


class AsyncAuthAPI(BaseAsyncAPI):
    async def account_info(
        self, address: AccAddress
    ) -> Union[
        BaseAccount,
        ContinuousVestingAccount,
        DelayedVestingAccount,
        PeriodicVestingAccount,
    ]:
        """Fetches the account information.

        Args:
            address (AccAddress): account address

        Returns:
            Union[BaseAccount, ContinuousVestingAccount, DelayedVestingAccount, PeriodicVestingAccount]: account information
        """
        result = await self._c._get(f"/cosmos/auth/v1beta1/accounts/{address}")
        return Account.from_data(result["account"])


class AuthAPI(AsyncAuthAPI):
    @sync_bind(AsyncAuthAPI.account_info)
    def account_info(
        self, address: AccAddress
    ) -> Union[
        BaseAccount,
        ContinuousVestingAccount,
        DelayedVestingAccount,
        PeriodicVestingAccount,
    ]:
        pass

    account_info.__doc__ = AsyncAuthAPI.account_info.__doc__



import abc
from abc import ABC
from typing import Optional

__all__ = ["APIParams", "PaginationOptions"]


class APIParams(ABC):
    @abc.abstractmethod
    def to_dict(self):
        pass

    def to_list(self) -> list:
        lst = []
        dct = self.to_dict()
        for key in dct.keys():
            lst.append((key, dct.get(key)))
        return lst


class PaginationOptions(APIParams):
    """This could be used when you need pagination options for APIs

    Args:
        key (str): key is a value returned in PageResponse.next_key to begin
            querying the next page most efficiently. Only one of offset or key
            should be set.
        offset (int): offset is a numeric offset that can be used when key is unavailable.
            It is less efficient than using key. Only one of offset or key should be set.
        limit (int): limit is the total number of results to be returned in the result page.
            If left empty it will default to a value to be set by each app.
        count_total (bool): count_total is set to true to indicate that the result set should include a count of
            the total number of items available for pagination in UIs.
            count_total is only respected when offset is used. It is ignored when key is set.
        reverse (bool): reverse is set to true if results are to be returned in the descending order.
    """

    def __init__(
        self,
        key: Optional[str] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        count_total: Optional[bool] = None,
        reverse: Optional[bool] = None,
    ):
        self.key = key
        self.offset = offset
        self.limit = limit
        self.count_total = count_total
        self.reverse = reverse

    def __str__(self):
        return "&".join(self.to_dict())

    def to_dict(self):
        params = {}
        if self.key is not None:
            params["pagination.key"] = self.key
        if self.offset is not None:
            params["pagination.offset"] = self.offset
        if self.limit is not None:
            params["pagination.limit"] = self.limit
        if self.count_total is not None:
            params["pagination.count_total"] = str(self.count_total).lower()
        if self.reverse is not None:
            params["pagination.reverse"] = str(self.reverse).lower()
        return params



"""Data objects pertaining to accounts."""

from __future__ import annotations

from itertools import starmap
from typing import List

import attr
from terra_proto.cosmos.vesting.v1beta1 import Period as Period_pb
from terra_proto.cosmos.vesting.v1beta1 import (
    PeriodicVestingAccount as PeriodicVestingAccount_pb,
)

from terra_sdk.core import Coins
from terra_sdk.util.json import JSONSerializable

from ...public_key import PublicKey
from .base_account import BaseAccount
from .base_vesting_account import BaseVestingAccount

__all__ = ["Period", "PeriodicVestingAccount"]


@attr.s
class Period(JSONSerializable):
    length: int = attr.ib(converter=int)
    amount: Coins = attr.ib(converter=Coins)

    def to_amino(self) -> dict:
        return {
            "length": str(self.length),
            "amount": self.amount.to_amino(),
        }

    def to_data(self) -> dict:
        return {
            "length": str(self.length),
            "amount": self.amount.to_data(),
        }

    def to_proto(self) -> Period_pb:
        return Period_pb(length=self.length, amount=self.amount.to_proto())

    @classmethod
    def from_amino(cls, amino: dict) -> Period:
        return cls(
            length=int(amino["length"]), amount=Coins.from_amino(amino["amount"])
        )

    @classmethod
    def from_data(cls, data: dict) -> Period:
        return cls(length=data["length"], amount=Coins.from_data(data["amount"]))

    @classmethod
    def from_proto(cls, proto: Period_pb) -> Period:
        return cls(length=proto.length, amount=Coins.from_proto(proto.amount))


@attr.s
class PeriodicVestingAccount:
    """Stores information about an account with periodic vesting."""

    base_vesting_account: BaseVestingAccount = attr.ib()
    start_time: int = attr.ib()
    vesting_periods: List[Period] = attr.ib()

    type_amino = "cosmos-sdk/PeriodicVestingAccount"
    type_url = "/cosmos.vesting.v1beta1.PeriodicVestingAccount"

    def get_sequence(self) -> int:
        return self.base_vesting_account.get_sequence()

    def get_account_number(self) -> int:
        return self.base_vesting_account.get_account_number()

    def get_public_key(self) -> PublicKey:
        return self.base_vesting_account.get_public_key()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "base_vesting_account": self.base_vesting_account.to_amino(),
                "start_time": str(self.start_time),
                "vesting_periods": [vp.to_amino() for vp in self.vesting_periods],
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "base_vesting_account": self.base_vesting_account.to_data(),
            "start_time": str(self.start_time),
            "vesting_periods": [vp.to_data() for vp in self.vesting_periods],
        }

    def to_proto(self) -> PeriodicVestingAccount_pb:
        return PeriodicVestingAccount_pb(
            base_vesting_account=self.base_vesting_account.to_proto(),
            start_time=self.start_time,
            vesting_periods=[vp.to_proto for vp in self.vesting_periods],
        )

    @classmethod
    def from_amino(cls, amino: dict) -> PeriodicVestingAccount:
        amino = amino["value"]
        return cls(
            base_vesting_account=BaseVestingAccount.from_amino(
                {
                    "type": BaseVestingAccount.type_amino,
                    "value": amino["base_vesting_account"],
                }
            ),
            start_time=amino["start_time"],
            vesting_periods=[Period.from_amino(vp) for vp in amino["vesting_periods"]],
        )

    @classmethod
    def from_data(cls, data: dict) -> PeriodicVestingAccount:
        return cls(
            base_vesting_account=BaseVestingAccount.from_data(
                data["base_vesting_account"]
            ),
            start_time=data["start_time"],
            vesting_periods=[Period.from_data(vp) for vp in data["vesting_periods"]],
        )

    @classmethod
    def from_proto(cls, proto: PeriodicVestingAccount_pb) -> PeriodicVestingAccount:
        return cls(
            base_vesting_account=BaseVestingAccount.from_proto(
                proto.base_vesting_account
            ),
            start_time=proto.start_time,
            vesting_periods=[Period.from_proto(vs) for vs in proto.vesting_periods],
        )



from .msgs import (
    MsgFundCommunityPool,
    MsgSetWithdrawAddress,
    MsgWithdrawDelegatorReward,
    MsgWithdrawValidatorCommission,
)
from .proposals import CommunityPoolSpendProposal

__all__ = [
    "MsgFundCommunityPool",
    "MsgSetWithdrawAddress",
    "MsgWithdrawDelegatorReward",
    "MsgWithdrawValidatorCommission",
    "CommunityPoolSpendProposal",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['terra_sdk',
 'terra_sdk.client',
 'terra_sdk.client.lcd',
 'terra_sdk.client.lcd.api',
 'terra_sdk.core',
 'terra_sdk.core.auth',
 'terra_sdk.core.auth.data',
 'terra_sdk.core.authz',
 'terra_sdk.core.bank',
 'terra_sdk.core.crisis',
 'terra_sdk.core.distribution',
 'terra_sdk.core.feegrant',
 'terra_sdk.core.gov',
 'terra_sdk.core.ibc',
 'terra_sdk.core.ibc.data',
 'terra_sdk.core.ibc.msgs',
 'terra_sdk.core.ibc.proposals',
 'terra_sdk.core.ibc_transfer',
 'terra_sdk.core.params',
 'terra_sdk.core.slashing',
 'terra_sdk.core.staking',
 'terra_sdk.core.staking.data',
 'terra_sdk.core.upgrade',
 'terra_sdk.core.upgrade.data',
 'terra_sdk.core.wasm',
 'terra_sdk.key',
 'terra_sdk.util']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'attrs>=21.4.0,<22.0.0',
 'bech32>=1.2.0,<2.0.0',
 'betterproto==2.0.0b4',
 'bip32utils>=0.3.post4,<0.4',
 'boltons>=21.0.0,<22.0.0',
 'ecdsa>=0.17.0,<0.18.0',
 'furl>=2.1.3,<3.0.0',
 'mnemonic>=0.19,<0.20',
 'nest-asyncio>=1.5.4,<2.0.0',
 'protobuf>=3.19.1,<4.0.0',
 'terra-proto==2.0.0',
 'wrapt>=1.13.3,<2.0.0']

setup_kwargs = {
    'name': 'terra-sdk',
    'version': '3.0.0',
    'description': 'The Python SDK for Terra',
    'long_description': '<br/>\n<br/>\n\n<div  align="center"> <p > <img src="https://raw.githubusercontent.com/terra-money/terra-sdk-python/main/docs/img/logo.png" width=500 alt="py-sdk-logo"></p>\n\nThe Python SDK for Terra\n<br/>\n\n<p><sub>(Unfamiliar with Terra?  <a href="https://docs.terra.money/">Check out the Terra Docs</a>)</sub></p>\n\n  <p > <img alt="GitHub" src="https://img.shields.io/github/license/terra-money/terra-sdk-python">\n<img alt="Python" src="https://img.shields.io/pypi/pyversions/terra-sdk">\n  <img alt="pip" src="https://img.shields.io/pypi/v/terra-sdk"></p>\n<p>\n  <a href="https://terra-money.github.io/terra.py/index.html"><strong>Explore the Docs »</strong></a>\n<br/>\n  <a href="https://pypi.org/project/terra-sdk/">PyPI Package</a>\n  ·\n  <a href="https://github.com/terra-money/terra.py">GitHub Repository</a>\n</p></div>\n\nThe Terra Software Development Kit (SDK) in Python is a simple library toolkit for building software that can interact with the Terra blockchain and provides simple abstractions over core data structures, serialization, key management, and API request generation.\n\n## Features\n\n- Written in Python with extensive support libraries\n- Versatile support for key management solutions\n- Exposes the Terra API through LCDClient\n\n<br/>\n\n# Table of Contents\n\n- [API Reference](#api-reference)\n- [Getting Started](#getting-started)\n  - [Requirements](#requirements)\n  - [Installation](#installation)\n  - [Dependencies](#dependencies)\n  - [Tests](#tests)\n  - [Code Quality](#code-quality)\n- [Usage Examples](#usage-examples)\n  - [Getting Blockchain Information](#getting-blockchain-information)\n    - [Async Usage](#async-usage)\n  - [Building and Signing Transactions](#building-and-signing-transactions)\n    - [Example Using a Wallet](#example-using-a-wallet-recommended)\n- [Contributing](#contributing)\n  - [Reporting an Issue](#reporting-an-issue)\n  - [Requesting a Feature](#requesting-a-feature)\n  - [Contributing Code](#contributing-code)\n  - [Documentation Contributions](#documentation-contributions)\n- [License](#license)\n\n<br/>\n\n# API Reference\n\nAn intricate reference to the APIs on the Terra SDK can be found <a href="https://terra-money.github.io/terra.py/index.html">here</a>.\n\n<br/>\n\n# Getting Started\n\nA walk-through of the steps to get started with the Terra SDK alongside a few use case examples are provided below. Alternatively, a tutorial video is also available <a href="https://www.youtube.com/watch?v=GfasBlJHKIg">here</a> as reference.\n\n## Requirements\n\nTerra SDK requires <a href="https://www.python.org/downloads/">Python v3.7+</a>.\n\n## Installation\n\n<sub>**NOTE:** _All code starting with a `$` is meant to run on your terminal (a bash prompt). All code starting with a `>>>` is meant to run in a python interpreter, like <a href="https://pypi.org/project/ipython/">ipython</a>._</sub>\n\nTerra SDK can be installed (preferably in a `virtual environment` from PyPI using `pip`) as follows:\n  \n```\n$ pip install -U terra_sdk\n```\n\n<sub>_You might have `pip3` installed instead of `pip`; proceed according to your own setup._<sub>\n  \n❗ If you want to communicate with Terra Classic, use terra-sdk==2.x\n  \n## Dependencies\n\nTerra SDK uses <a href="https://python-poetry.org/">Poetry</a> to manage dependencies. To get set up with all the required dependencies, run:\n\n```\n$ pip install poetry\n$ poetry install\n```\n\n## Tests\n\nTerra SDK provides extensive tests for data classes and functions. To run them, after the steps in [Dependencies](#dependencies):\n\n```\n$ make test\n```\n\n## Code Quality\n\nTerra SDK uses <a href="https://black.readthedocs.io/en/stable/">Black</a>, <a href="https://isort.readthedocs.io/en/latest/">isort</a>, and <a href="https://mypy.readthedocs.io/en/stable/index.html">Mypy</a> for checking code quality and maintaining style. To reformat, after the steps in [Dependencies](#dependencies):\n\n```\n$ make qa && make format\n```\n\n<br/>\n\n# Usage Examples\n\nTerra SDK can help you read block data, sign and send transactions, deploy and interact with contracts, and many more.\nThe following examples are provided to help you get started. Use cases and functionalities of the Terra SDK are not limited to the following examples and can be found in full <a href="https://terra-money.github.io/terra.py/index.html">here</a>.\n\nIn order to interact with the Terra blockchain, you\'ll need a connection to a Terra node. This can be done through setting up an LCDClient (The LCDClient is an object representing an HTTP connection to a Terra LCD node.):\n\n```\n>>> from terra_sdk.client.lcd import LCDClient\n>>> terra = LCDClient(chain_id="phoenix-1", url="https://phoenix-lcd.terra.dev")\n```\n\n## Getting Blockchain Information\n\nOnce properly configured, the `LCDClient` instance will allow you to interact with the Terra blockchain. Try getting the latest block height:\n\n```\n>>> terra.tendermint.block_info()[\'block\'][\'header\'][\'height\']\n```\n\n`\'1687543\'`\n\n### Async Usage\n\nIf you want to make asynchronous, non-blocking LCD requests, you can use AsyncLCDClient. The interface is similar to LCDClient, except the module and wallet API functions must be awaited.\n\n<pre><code>\n>>> import asyncio \n>>> from terra_sdk.client.lcd import AsyncLCDClient\n\n>>> async def main():\n      <strong>terra = AsyncLCDClient("https://phoenix-lcd.terra.dev", "phoenix-1")</strong>\n      total_supply = await terra.bank.total()\n      print(total_supply)\n      <strong>await terra.session.close # you must close the session</strong>\n\n>>> asyncio.get_event_loop().run_until_complete(main())\n</code></pre>\n\n## Building and Signing Transactions\n\nIf you wish to perform a state-changing operation on the Terra blockchain such as sending tokens, swapping assets, withdrawing rewards, or even invoking functions on smart contracts, you must create a **transaction** and broadcast it to the network.\nTerra SDK provides functions that help create StdTx objects.\n\n### Example Using a Wallet (_recommended_)\n\nA `Wallet` allows you to create and sign a transaction in a single step by automatically fetching the latest information from the blockchain (chain ID, account number, sequence).\n\nUse `LCDClient.wallet()` to create a Wallet from any Key instance. The Key provided should correspond to the account you intend to sign the transaction with.\n  \n<sub>**NOTE:** *If you are using MacOS and got an exception \'bad key length\' from MnemonicKey, please check your python implementation. if `python3 -c "import ssl; print(ssl.OPENSSL_VERSION)"` returns LibreSSL 2.8.3, you need to reinstall python via pyenv or homebrew.*</sub>\n\n```\n>>> from terra_sdk.client.lcd import LCDClient\n>>> from terra_sdk.key.mnemonic import MnemonicKey\n\n>>> mk = MnemonicKey(mnemonic=MNEMONIC)\n>>> terra = LCDClient("https://phoenix-lcd.terra.dev", "phoenix-1")\n>>> wallet = terra.wallet(mk)\n```\n\nOnce you have your Wallet, you can simply create a StdTx using `Wallet.create_and_sign_tx`.\n\n```\n>>> from terra_sdk.core.fee import Fee\n>>> from terra_sdk.core.bank import MsgSend\n>>> from terra_sdk.client.lcd.api.tx import CreateTxOptions\n\n>>> tx = wallet.create_and_sign_tx(CreateTxOptions(\n        msgs=[MsgSend(\n            wallet.key.acc_address,\n            RECIPIENT,\n            "1000000uluna"    # send 1 luna\n        )],\n        memo="test transaction!",\n        fee=Fee(200000, "120000uluna")\n    ))\n```\n\nYou should now be able to broadcast your transaction to the network.\n\n```\n>>> result = terra.tx.broadcast(tx)\n>>> print(result)\n```\n\n<br/>\n\n# Contributing\n\nCommunity contribution, whether it\'s a new feature, correction, bug report, additional documentation, or any other feedback is always welcome. Please read through this section to ensure that your contribution is in the most suitable format for us to effectively process.\n\n<br/>\n\n## Reporting an Issue\n\nFirst things first: **Do NOT report security vulnerabilities in public issues!** Please disclose responsibly by submitting your findings to the [Terra Bugcrowd submission form](https://www.terra.money/bugcrowd). The issue will be assessed as soon as possible.\nIf you encounter a different issue with the Python SDK, check first to see if there is an existing issue on the <a href="https://github.com/terra-money/terra-sdk-python/issues">Issues</a> page, or if there is a pull request on the <a href="https://github.com/terra-money/terra-sdk-python/pulls">Pull requests</a> page. Be sure to check both the Open and Closed tabs addressing the issue.\n\nIf there isn\'t a discussion on the topic there, you can file an issue. The ideal report includes:\n\n- A description of the problem / suggestion.\n- How to recreate the bug.\n- If relevant, including the versions of your:\n  - Python interpreter\n  - Terra SDK\n  - Optionally of the other dependencies involved\n- If possible, create a pull request with a (failing) test case demonstrating what\'s wrong. This makes the process for fixing bugs quicker & gets issues resolved sooner.\n  </br>\n\n## Requesting a Feature\n\nIf you wish to request the addition of a feature, please first check out the <a href="https://github.com/terra-money/terra-sdk-python/issues">Issues</a> page and the <a href="https://github.com/terra-money/terra-sdk-python/pulls">Pull requests</a> page (both Open and Closed tabs). If you decide to continue with the request, think of the merits of the feature to convince the project\'s developers, and provide as much detail and context as possible in the form of filing an issue on the <a href="https://github.com/terra-money/terra-sdk-python/issues">Issues</a> page.\n\n<br/>\n\n## Contributing Code\n\nIf you wish to contribute to the repository in the form of patches, improvements, new features, etc., first scale the contribution. If it is a major development, like implementing a feature, it is recommended that you consult with the developers of the project before starting the development to avoid duplicating efforts. Once confirmed, you are welcome to submit your pull request.\n</br>\n\n### For new contributors, here is a quick guide:\n\n1. Fork the repository.\n2. Build the project using the [Dependencies](#dependencies) and [Tests](#tests) steps.\n3. Install a <a href="https://virtualenv.pypa.io/en/latest/index.html">virtualenv</a>.\n4. Develop your code and test the changes using the [Tests](#tests) and [Code Quality](#code-quality) steps.\n5. Commit your changes (ideally follow the <a href="https://github.com/angular/angular/blob/master/CONTRIBUTING.md#commit">Angular commit message guidelines</a>).\n6. Push your fork and submit a pull request to the repository\'s `main` branch to propose your code.\n\nA good pull request:\n\n- Is clear and concise.\n- Works across all supported versions of Python. (3.7+)\n- Follows the existing style of the code base (<a href="https://pypi.org/project/flake8/">`Flake8`</a>).\n- Has comments included as needed.\n- Includes a test case that demonstrates the previous flaw that now passes with the included patch, or demonstrates the newly added feature.\n- Must include documentation for changing or adding any public APIs.\n- Must be appropriately licensed (MIT License).\n  </br>\n\n## Documentation Contributions\n\nDocumentation improvements are always welcome. The documentation files live in the [docs](./docs) directory of the repository and are written in <a href="https://docutils.sourceforge.io/rst.html">reStructuredText</a> and use <a href="https://www.sphinx-doc.org/en/master/">Sphinx</a> to create the full suite of documentation.\n</br>\nWhen contributing documentation, please do your best to follow the style of the documentation files. This means a soft limit of 88 characters wide in your text files and a semi-formal, yet friendly and approachable, prose style. You can propose your improvements by submitting a pull request as explained above.\n\n### Need more information on how to contribute?\n\nYou can give this <a href="https://opensource.guide/how-to-contribute/#how-to-submit-a-contribution">guide</a> read for more insight.\n\n<br/>\n\n# License\n\nThis software is licensed under the MIT license. See [LICENSE](./LICENSE) for full disclosure.\n\n© 2021 Terraform Labs, PTE.\n\n<hr/>\n\n<p>&nbsp;</p>\n<p align="center">\n    <a href="https://terra.money/"><img src="https://assets.website-files.com/611153e7af981472d8da199c/61794f2b6b1c7a1cb9444489_symbol-terra-blue.svg" alt="Terra-logo" width=200/></a>\n<div align="center">\n  <sub><em>Powering the innovation of money.</em></sub>\n</div>\n',
    'author': 'Terraform Labs, PTE.',
    'author_email': 'engineering@terra.money',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/terra-money/terra.py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from .delegation import (
    Delegation,
    Redelegation,
    RedelegationEntry,
    UnbondingDelegation,
    UnbondingDelegationEntry,
)
from .validator import Commission, CommissionRates, Description, Validator

__all__ = [
    "Delegation",
    "UnbondingDelegation",
    "UnbondingDelegationEntry",
    "Redelegation",
    "RedelegationEntry",
    "CommissionRates",
    "Commission",
    "Description",
    "Validator",
]



from __future__ import annotations

from betterproto.lib.google.protobuf import Any as Any_pb

from terra_sdk.util.base import BaseTerraData


class Msg(BaseTerraData):
    def to_proto(self):
        raise NotImplementedError

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    @staticmethod
    def from_data(data: dict) -> Msg:
        from terra_sdk.util.parse_msg import parse_msg

        return parse_msg(data)

    @staticmethod
    def from_proto(proto: any) -> Msg:
        from terra_sdk.util.parse_msg import parse_proto

        return parse_proto(proto)

    @staticmethod
    def unpack_any(any_pb: Any_pb) -> Msg:
        from terra_sdk.util.parse_msg import parse_unpack_any

        return parse_unpack_any(any_pb)



from __future__ import annotations

from datetime import datetime
from typing import List

import attr
from dateutil import parser
from terra_proto.cosmos.staking.v1beta1 import Delegation as Delegation_pb
from terra_proto.cosmos.staking.v1beta1 import (
    DelegationResponse as DelegationResponse_pb,
)
from terra_proto.cosmos.staking.v1beta1 import Redelegation as Redelegation_pb
from terra_proto.cosmos.staking.v1beta1 import RedelegationEntry as RedelegationEntry_pb
from terra_proto.cosmos.staking.v1beta1 import (
    RedelegationEntryResponse as RedelegationEntryResponse_pb,
)
from terra_proto.cosmos.staking.v1beta1 import (
    UnbondingDelegation as UnbondingDelegation_pb,
)
from terra_proto.cosmos.staking.v1beta1 import (
    UnbondingDelegationEntry as UnbondingDelegationEntry_pb,
)

from terra_sdk.core import AccAddress, Coin, Dec, ValAddress
from terra_sdk.util.converter import to_isoformat
from terra_sdk.util.json import JSONSerializable

__all__ = [
    "Delegation",
    "UnbondingDelegation",
    "UnbondingDelegationEntry",
    "Redelegation",
    "RedelegationEntry",
]


@attr.s
class DelegationInfo(JSONSerializable):
    delegator_address: AccAddress = attr.ib()
    validator_address: ValAddress = attr.ib()
    shares: Dec = attr.ib(converter=Dec)


@attr.s
class Delegation(JSONSerializable):
    """Contains information about a current delegation pair (``delegator_address``, ``validator_address``)"""

    delegation: DelegationInfo = attr.ib()
    balance: Coin = attr.ib(converter=Coin.parse)  # type: ignore

    def to_amino(self) -> dict:
        return {
            "delegation": {
                "delegator_address": self.delegation.delegator_address,
                "validator_address": self.delegation.validator_address,
                "shares": str(self.delegation.shares),
            },
            "balance": self.balance.to_amino(),
        }

    @classmethod
    def from_data(cls, data: dict) -> Delegation:
        return cls(
            delegation=DelegationInfo(
                delegator_address=data["delegation"]["delegator_address"],
                validator_address=data["delegation"]["validator_address"],
                shares=data["delegation"]["shares"],
            ),
            balance=Coin.from_data(data["balance"]),
        )

    def to_proto(self) -> DelegationResponse_pb:
        return DelegationResponse_pb(
            delegation=Delegation_pb(
                delegator_address=self.delegation.delegator_address,
                validator_address=self.delegation.validator_address,
                shares=str(self.delegation.shares),
            ),
            balance=self.balance.to_proto(),
        )


@attr.s
class UnbondingDelegationEntry(JSONSerializable):
    """Contains information about an active unbonding lot of Luna."""

    initial_balance: int = attr.ib(converter=int)
    """"""
    balance: int = attr.ib(converter=int)
    """"""
    creation_height: int = attr.ib(converter=int)
    """"""
    completion_time: datetime = attr.ib(converter=parser.parse)
    """"""

    def to_amino(self) -> dict:
        return {
            "initial_balance": str(self.initial_balance),
            "balance": str(self.balance),
            "creation_height": str(self.creation_height),
            "completion_time": to_isoformat(self.completion_time),
        }

    def to_data(self) -> dict:
        return {
            "initial_balance": str(self.initial_balance),
            "balance": str(self.balance),
            "creation_height": str(self.creation_height),
            "completion_time": to_isoformat(self.completion_time),
        }

    @classmethod
    def from_data(cls, data: dict) -> UnbondingDelegationEntry:
        return cls(
            initial_balance=data["initial_balance"],
            balance=data["balance"],
            creation_height=data["creation_height"],
            completion_time=data["completion_time"],
        )

    def to_proto(self) -> UnbondingDelegationEntry_pb:
        return UnbondingDelegationEntry_pb(
            initial_balance=str(self.initial_balance),
            balance=str(self.balance),
            creation_height=self.creation_height,
            completion_time=self.completion_time,
        )


@attr.s
class UnbondingDelegation(JSONSerializable):
    """Contains information about undelegations for a delegation pair (``delegator_address``, ``validator_address``)"""

    delegator_address: AccAddress = attr.ib()
    """"""
    validator_address: ValAddress = attr.ib()
    """"""
    entries: List[UnbondingDelegationEntry] = attr.ib()
    """"""

    def to_amino(self) -> dict:
        return {
            "delegator_address": self.delegator_address,
            "validator_address": self.validator_address,
            "entries": [entry.to_amino() for entry in self.entries],
        }

    @classmethod
    def from_data(cls, data: dict) -> UnbondingDelegation:
        entries = [
            UnbondingDelegationEntry.from_data(entry) for entry in data["entries"]
        ]
        return cls(
            delegator_address=data["delegator_address"],
            validator_address=data["validator_address"],
            entries=entries,
        )

    def to_proto(self) -> UnbondingDelegation_pb:
        return UnbondingDelegation_pb(
            delegator_address=self.delegator_address,
            validator_address=self.validator_address,
            entries=[entry.to_proto() for entry in self.entries],
        )


@attr.s
class RedelegationEntryInfo(JSONSerializable):
    initial_balance: int = attr.ib(converter=int)
    """"""
    shares_dst: Dec = attr.ib(converter=Dec)
    """"""
    creation_height: int = attr.ib(converter=int)
    """"""
    completion_time: datetime = attr.ib(converter=parser.parse)
    """"""

    def to_amino(self) -> dict:
        return {
            "initial_balance": str(self.initial_balance),
            "shares_dst": str(self.shares_dst),
            "creation_height": str(self.creation_height),
            "completion_time": to_isoformat(self.completion_time),
        }

    @classmethod
    def from_data(cls, data: dict) -> RedelegationEntryInfo:
        return cls(
            initial_balance=data["initial_balance"],
            shares_dst=Dec.from_data(data("shares_dst")),
            creation_height=data["creation_height"],
            completion_time=data["completion_time"],
        )

    def to_data(self) -> dict:
        return {
            "initial_balance": self.initial_balance,
            "shares_dst": self.shares_dst.to_data(),
            "creation_height": self.creation_height,
            "completion_time": to_isoformat(self.completion_time),
        }

    def to_proto(self) -> RedelegationEntry_pb:
        return RedelegationEntry_pb(
            initial_balance=str(self.initial_balance),
            shares_dst=str(self.shares_dst),
            creation_height=self.creation_height,
            completion_time=self.completion_time,
        )


@attr.s
class RedelegationEntry(JSONSerializable):
    """Contains information about an active redelegated lot of Luna."""

    redelegation_entry: RedelegationEntryInfo = attr.ib()
    """"""
    balance: int = attr.ib(converter=int)
    """"""

    def to_amino(self) -> dict:
        return {
            "redelegation_entry": self.redelegation_entry.to_amino(),
            "balance": str(self.balance),
        }

    def to_data(self) -> dict:
        return {
            "redelegation_entry": {
                "initial_balance": str(self.redelegation_entry.initial_balance),
                "shares_dst": str(self.redelegation_entry.shares_dst),
                "creation_height": self.redelegation_entry.creation_height,
                "completion_time": self.redelegation_entry.completion_time,
            },
            "balance": str(self.balance),
        }

    @classmethod
    def from_data(cls, data: dict) -> RedelegationEntry:
        return cls(
            redelegation_entry=RedelegationEntryInfo(
                initial_balance=data["redelegation_entry"]["initial_balance"],
                shares_dst=data["redelegation_entry"]["shares_dst"],
                creation_height=int(data["redelegation_entry"]["creation_height"]),
                completion_time=data["redelegation_entry"]["completion_time"],
            ),
            balance=data["balance"],
        )

    def to_proto(self) -> RedelegationEntryResponse_pb:
        return RedelegationEntryResponse_pb(
            redelegation_entry=self.redelegation_entry.to_proto(),
            balance=str(self.balance),
        )


@attr.s
class RedelegationInfo(JSONSerializable):
    delegator_address: AccAddress = attr.ib()
    """"""
    validator_src_address: ValAddress = attr.ib()
    """"""
    validator_dst_address: ValAddress = attr.ib()
    """"""

    def to_amino(self) -> dict:
        return {
            "delegator_address": self.delegator_address,
            "validator_src_address": self.validator_src_address,
            "validator_dst_address": self.validator_dst_address,
        }

    def to_data(self) -> dict:
        return {
            "delegator_address": self.delegator_address,
            "validator_src_address": self.validator_src_address,
            "validator_dst_address": self.validator_dst_address,
        }

    @classmethod
    def from_data(cls, data: dict) -> RedelegationInfo:
        return cls(
            delegator_address=data["delegator_address"],
            validator_src_address=data["validator_src_address"],
            validator_dst_address=data["validator_dst_address"],
        )

    def to_proto(self) -> Redelegation_pb:
        return Redelegation_pb(
            delegator_address=self.delegator_address,
            validator_src_address=self.validator_src_address,
            validator_dst_address=self.validator_dst_address,
        )


@attr.s
class Redelegation(JSONSerializable):
    """Contains informations about a redelgation for delegation tuple (``delegator_address``, ``validator_src_address``, ``validator_dst_address``)"""

    redelegation: RedelegationInfo = attr.ib()
    """"""
    entries: List[RedelegationEntry] = attr.ib()
    """"""

    def to_amino(self) -> dict:
        return {
            "redelegation": self.redelegation.to_amino(),
            "entries": [entry.to_amino() for entry in self.entries],
        }

    @classmethod
    def from_data(cls, data: dict) -> Redelegation:
        entries = [RedelegationEntry.from_data(re) for re in data["entries"]]
        return cls(
            redelegation=RedelegationInfo(
                delegator_address=data["redelegation"]["delegator_address"],
                validator_src_address=data["redelegation"]["validator_src_address"],
                validator_dst_address=data["redelegation"]["validator_dst_address"],
            ),
            entries=entries,
        )

    def to_proto(self) -> Redelegation_pb:
        return Redelegation_pb(
            delegator_address=self.redelegation.delegator_address,
            validator_src_address=self.redelegation.validator_src_address,
            validator_dst_address=self.redelegation.validator_dst_address,
            entries=[entry.to_proto() for entry in self.entries],
        )



from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncIbcAPI", "IbcAPI"]


class AsyncIbcAPI(BaseAsyncAPI):
    async def parameters(self) -> dict:
        """Fetches the Ibc module's parameters.

        Returns:
            List: allowed clients
        """
        res = await self._c._get("/ibc/client/v1/params")
        return res["params"]

    # TODO: functions for clients, connections and channels


class IbcAPI(AsyncIbcAPI):
    @sync_bind(AsyncIbcAPI.parameters)
    def parameters(self) -> dict:
        pass

    parameters.__doc__ = AsyncIbcAPI.parameters.__doc__



"""CompactBitArray types related to multisig."""
from __future__ import annotations

import base64
import math

import attr
from terra_proto.cosmos.crypto.multisig.v1beta1 import (
    CompactBitArray as CompactBitArray_pb,
)

from terra_sdk.util.json import JSONSerializable

__all__ = ["CompactBitArray"]


@attr.s
class CompactBitArray(JSONSerializable):
    extra_bits_stored: int = attr.ib(converter=int)
    elems: bytearray = attr.ib(converter=bytearray)

    @classmethod
    def from_data(cls, data: dict) -> CompactBitArray:
        return cls(
            data["extra_bits_stored"], bytearray(base64.b64decode(data["elems"]))
        )

    def to_data(self) -> dict:
        return {
            "extra_bits_stored": self.extra_bits_stored,
            "elems": base64.b64encode(self.elems),
        }

    @classmethod
    def from_proto(cls, proto: CompactBitArray_pb) -> CompactBitArray:
        return cls(proto.extra_bits_stored, bytearray(proto.elems))

    def to_proto(self) -> CompactBitArray_pb:
        return CompactBitArray_pb(
            extra_bits_stored=self.extra_bits_stored, elems=bytes(self.elems)
        )

    @classmethod
    def from_bits(cls, bits: int) -> CompactBitArray:
        if bits <= 0:
            raise ValueError("CompactBitArray bits must be bigger than 0")

        num_elems = (bits + 7) // 8
        if num_elems <= 0 or num_elems > (math.pow(2, 32) - 1):
            raise ValueError("CompactBitArray overflow")

        return CompactBitArray(bits % 8, bytearray(num_elems))

    def count(self) -> int:
        if self.extra_bits_stored == 0:
            return len(self.elems) * 8
        return (len(self.elems) - 1) * 8 + self.extra_bits_stored

    def get_index(self, i: int) -> bool:
        if i < 0 or i >= self.count():
            return False
        return (self.elems[(i >> 3)] & (1 << (7 - (i % 8)))) > 0

    def set_index(self, i: int, v: bool) -> bool:
        if i < 0 or i >= self.count():
            return False
        if v:  # True
            self.elems[i >> 3] |= 1 << (7 - (i % 8))
        else:  # False
            self.elems[i >> 3] &= ~(1 << (7 - (i % 8)))
        return True

    def num_true_bits_before(self, index: int) -> int:
        def count_one_bits(n: int):
            return len("".join("{0:b}".format(n).split("0")))

        ones_count = 0
        _max = self.count()
        if index > _max:
            index = _max

        elem = 0
        while True:
            if (elem * 8 + 7) >= index:
                ones_count += count_one_bits(self.elems[elem] >> (7 - (index % 8) + 1))
                return ones_count
            ones_count += count_one_bits(self.elems[elem])
            elem += 1



from typing import List, Optional

import attr

from terra_sdk.core import AccAddress, Coin, Numeric, ValAddress
from terra_sdk.core.staking import (
    Delegation,
    Redelegation,
    UnbondingDelegation,
    Validator,
)

from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncStakingAPI", "StakingAPI", "StakingPool"]

from ..params import APIParams, PaginationOptions


class RedelegationsOptions(PaginationOptions):
    """just internal class for relegation option"""

    def __init__(
        self,
        src_validator_addr: Optional[str] = None,
        dst_validator_addr: Optional[str] = None,
    ):
        super().__init__(self)
        self.src_validator_addr = src_validator_addr
        self.dst_validator_addr = dst_validator_addr

    def __str__(self):
        return "&".join(self.to_dict())

    def to_dict(self) -> dict:
        params = super().to_dict()
        if self.src_validator_addr is not None:
            params["src_validator_addr"] = self.src_validator_addr
        if self.dst_validator_addr is not None:
            params["dst_validator_addr"] = self.dst_validator_addr
        return params


@attr.s
class StakingPool:
    bonded_tokens: Coin = attr.ib()
    not_bonded_tokens: Coin = attr.ib()


class AsyncStakingAPI(BaseAsyncAPI):
    async def delegations(
        self,
        delegator: Optional[AccAddress] = None,
        validator: Optional[ValAddress] = None,
        params: Optional[APIParams] = None,
    ) -> (List[Delegation], dict):
        """Fetches current delegations, filtering by delegator, validator, or both.

        Args:
            delegator (Optional[AccAddress], optional): delegator account address.
            validator (Optional[ValAddress], optional): validator operator address.
            params (APIParams, optional): additional params for the API like pagination

        Raises:
            TypeError: if both ``delegator`` and ``validator`` are ``None``.

        Returns:
            List[Delegation]: delegations
            dict: pagination info
        """
        if delegator is not None and validator is not None:
            res = await self._c._get(
                f"/cosmos/staking/v1beta1/validators/{validator}/delegations/{delegator}",
                params,
            )
            return [Delegation.from_data(res.get("delegation_response"))], res.get(
                "pagination"
            )
        elif delegator is not None:
            res = await self._c._get(
                f"/cosmos/staking/v1beta1/delegations/{delegator}", params
            )
            return [
                Delegation.from_data(d) for d in res.get("delegation_responses")
            ], res.get("pagination")
        elif validator is not None:
            res = await self._c._get(
                f"/cosmos/staking/v1beta1/validators/{validator}/delegations", params
            )
            return [
                Delegation.from_data(d) for d in res.get("delegation_responses")
            ], res.get("pagination")
        else:
            raise TypeError("arguments delegator and validator cannot both be None")

    async def delegation(
        self, delegator: AccAddress, validator: ValAddress
    ) -> Delegation:
        """Fetch a single delegation via a delegator, validator pair.

        Args:
            delegator (Optional[AccAddress), optional: delegator account address
            validator (Optional[ValAddress], optional): validator operator address

        Returns:
            Delegation: delegation
        """
        res = await self._c._get(
            f"/cosmos/staking/v1beta1/validators/{validator}/delegations/{delegator}"
        )
        res = res.get("delegation_response").get("delegation")
        return res

    async def unbonding_delegations(
        self,
        delegator: Optional[AccAddress] = None,
        validator: Optional[ValAddress] = None,
        params: Optional[APIParams] = None,
    ) -> (List[UnbondingDelegation], dict):
        """Fetches current undelegations, filtering by delegator, validator, or both.

        Args:
            delegator (Optional[AccAddress], optional): delegator account address.
            validator (Optional[ValAddress], optional): validator operator address.
            params (APIParams, optional): additional params for the API like pagination

        Raises:
            TypeError: if both ``delegator`` and ``validator`` are ``None``.

        Returns:
            List[UnbondingDelegation]: undelegations
            dict: pagination info
        """
        if delegator is not None and validator is not None:
            res = await self._c._get(
                f"/cosmos/staking/v1beta1/validators/{validator}/delegations/{delegator}/unbonding_delegation",
                params,
            )
            return [UnbondingDelegation.from_data(res.get("unbond"))], res.get(
                "pagination"
            )
        elif delegator is not None:
            res = await self._c._get(
                f"/cosmos/staking/v1beta1/delegators/{delegator}/unbonding_delegations",
                params,
            )
            return [
                UnbondingDelegation.from_data(x) for x in res.get("unbonding_responses")
            ], res.get("pagination")
        elif validator is not None:
            res = await self._c._get(
                f"/cosmos/staking/v1beta1/validators/{validator}/unbonding_delegations",
                params,
            )
            return [
                UnbondingDelegation.from_data(x) for x in res.get("unbonding_responses")
            ], res.get("pagination")
        else:
            raise TypeError("arguments delegator and validator cannot both be None")

    async def unbonding_delegation(
        self, delegator: AccAddress, validator: ValAddress
    ) -> UnbondingDelegation:
        """Fetch a single undelegation via a delegator, validator pair.

        Args:
            delegator (AccAddress): delegator account address
            validator (ValAddress): validator operator address

        Returns:
            UnbondingDelegation: undelegation
        """
        res = await self._c._get(
            f"/cosmos/staking/v1beta1/validators/{validator}/delegations/{delegator}/unbonding_delegation"
        )
        return UnbondingDelegation.from_data(res.get("unbond"))

    async def redelegations(
        self,
        delegator: Optional[AccAddress] = None,
        validator_src: Optional[ValAddress] = None,
        validator_dst: Optional[ValAddress] = None,
        params: Optional[APIParams] = None,
    ) -> (List[Redelegation], dict):
        """Fetch redelgations.

        Args:
            delegator (Optional[AccAddress], optional): delegator account address.
            validator_src (Optional[ValAddress], optional): source validator operator address (from).
            validator_dst (Optional[ValAddress], optional): dest. validator operator address (to).
            params (APIParams, optional): additional params for the API like pagination

        Returns:
            List[Redelegation]: redelegations
            dict: pagination info
        """

        # _params = RedelegationsOptions(src_validator_addr=validator_src, dst_validator_addr=validator_dst)
        if params is not None:
            _params = params.to_dict()
        else:
            _params = {}
        _params["src_validator_addr"] = validator_src
        _params["dst_validator_addr"] = validator_dst
        for x in list(_params.keys()):
            if _params[x] is None:
                del _params[x]
        res = await self._c._get(
            f"/cosmos/staking/v1beta1/delegators/{delegator}/redelegations", _params
        )
        return [
            Redelegation.from_data(d) for d in res.get("redelegation_responses")
        ], res.get("pagination")

    async def bonded_validators(
        self, delegator: AccAddress, params: Optional[PaginationOptions]
    ) -> (List[Validator], dict):
        """Fetches the list of validators a delegator is currently delegating to.

        Args:
            delegator (AccAddress): delegator account address
            params (APIParams, optional): additional params for the API like pagination

        Returns:
            List[Validator]: currently bonded validators
            dict: pagination info
        """
        res = await self._c._get(
            f"/cosmos/staking/v1beta1/delegators/{delegator}/validators", params
        )
        return [Validator.from_data(d) for d in res.get("validators")], res.get(
            "pagination"
        )

    async def validators(
        self, params: Optional[APIParams] = None
    ) -> (List[Validator], dict):
        """Fetch information of all validators.

        Args:
            params (APIParams, optional): additional params for the API like pagination

        Returns:
            List[Validator]: validator informations
            dict: pagination info
        """
        res = await self._c._get("/cosmos/staking/v1beta1/validators", params)
        return [Validator.from_data(d) for d in res.get("validators")], res.get(
            "pagination"
        )

    async def validator(self, validator: ValAddress) -> Validator:
        """Fetch information about a single validator.

        Args:
            validator (ValAddress): validator operator address

        Returns:
            Validator: validator information
        """
        res = await self._c._get(f"/cosmos/staking/v1beta1/validators/{validator}")
        return Validator.from_data(res.get("validator"))

    async def pool(self) -> StakingPool:
        """Fetch current staking pool information.

        Returns:
            StakingPool: information about current staking pool
        """
        res = await self._c._get("/cosmos/staking/v1beta1/pool")
        res = res.get("pool")
        return StakingPool(
            bonded_tokens=Coin("uluna", res["bonded_tokens"]),
            not_bonded_tokens=Coin("uluna", res["not_bonded_tokens"]),
        )

    async def parameters(self) -> dict:
        """Fetch Staking module parameters.

        Returns:
            dict: Staking module parameters
        """
        res = await self._c._get("/cosmos/staking/v1beta1/params")
        res = res.get("params")
        return {
            "unbonding_time": res["unbonding_time"],
            "max_validators": Numeric.parse(res["max_validators"]),
            "max_entries": Numeric.parse(res["max_entries"]),
            "historical_entries": Numeric.parse(res["historical_entries"]),
            "bond_denom": res["bond_denom"],
        }


class StakingAPI(AsyncStakingAPI):
    @sync_bind(AsyncStakingAPI.delegations)
    def delegations(
        self,
        delegator: Optional[AccAddress] = None,
        validator: Optional[ValAddress] = None,
        params: Optional[APIParams] = None,
    ) -> (List[Delegation], dict):
        pass

    delegations.__doc__ = AsyncStakingAPI.delegations.__doc__

    @sync_bind(AsyncStakingAPI.delegation)
    def delegation(self, delegator: AccAddress, validator: ValAddress) -> Delegation:
        pass

    delegation.__doc__ = AsyncStakingAPI.delegation.__doc__

    @sync_bind(AsyncStakingAPI.unbonding_delegations)
    def unbonding_delegations(
        self,
        delegator: Optional[AccAddress] = None,
        validator: Optional[ValAddress] = None,
        params: Optional[APIParams] = None,
    ) -> (List[UnbondingDelegation], dict):
        pass

    unbonding_delegations.__doc__ = AsyncStakingAPI.unbonding_delegations.__doc__

    @sync_bind(AsyncStakingAPI.unbonding_delegation)
    def unbonding_delegation(
        self, delegator: AccAddress, validator: ValAddress
    ) -> UnbondingDelegation:
        pass

    unbonding_delegation.__doc__ = AsyncStakingAPI.unbonding_delegation.__doc__

    @sync_bind(AsyncStakingAPI.redelegations)
    def redelegations(
        self,
        delegator: Optional[AccAddress] = None,
        validator_src: Optional[ValAddress] = None,
        validator_dst: Optional[ValAddress] = None,
        params: Optional[APIParams] = None,
    ) -> (List[Redelegation], dict):
        pass

    redelegations.__doc__ = AsyncStakingAPI.redelegations.__doc__

    @sync_bind(AsyncStakingAPI.bonded_validators)
    def bonded_validators(
        self, delegator: AccAddress, params: Optional[PaginationOptions] = None
    ) -> (List[Validator], dict):
        pass

    bonded_validators.__doc__ = AsyncStakingAPI.bonded_validators.__doc__

    @sync_bind(AsyncStakingAPI.validators)
    def validators(self, params: Optional[APIParams]) -> (List[Validator], dict):
        pass

    validators.__doc__ = AsyncStakingAPI.validators.__doc__

    @sync_bind(AsyncStakingAPI.validator)
    def validator(self, validator: ValAddress) -> Validator:
        pass

    validator.__doc__ = AsyncStakingAPI.validator.__doc__

    @sync_bind(AsyncStakingAPI.pool)
    def pool(self) -> StakingPool:
        pass

    pool.__doc__ = AsyncStakingAPI.pool.__doc__

    @sync_bind(AsyncStakingAPI.parameters)
    def parameters(self) -> dict:
        pass

    parameters.__doc__ = AsyncStakingAPI.parameters.__doc__



from __future__ import annotations

import math
import re
from typing import Union

import attr
from terra_proto.cosmos.base.v1beta1 import Coin as Coin_pb

from terra_sdk.util.json import JSONSerializable

from .numeric import Dec, Numeric


@attr.s(frozen=True)
class Coin(JSONSerializable):
    """Represents a (denom, amount) pairing, analagous to ``sdk.Coin`` and ``sdk.DecCoin``
    in Cosmos SDK. Used for representing Terra native assets.
    """

    denom: str = attr.ib()
    """Coin's denomination, only ``uluna``."""

    amount: Numeric.Output = attr.ib(converter=Numeric.parse)  # type: ignore
    """Coin's amount -- can be a ``int`` or :class:`Dec`"""

    @staticmethod
    def parse(arg: Union[Coin, str, dict]) -> Coin:
        """Converts the argument into a coin.

        Args:
            arg (Union[Coin, str, dict]): value to be converted to coin
        """
        if isinstance(arg, Coin):
            return arg
        elif isinstance(arg, str):
            return Coin.from_str(arg)
        else:
            return Coin.from_data(arg)

    def is_int_coin(self) -> bool:
        """Checks whether the coin's amount is of type ``int``."""
        return isinstance(self.amount, int)

    def is_dec_coin(self) -> bool:
        """Checks whether the coin's amount is of type :class:`Dec`."""
        return isinstance(self.amount, Dec)

    def to_int_coin(self) -> Coin:
        """Creates a new :class:`Coin` with an ``int`` amount."""
        return Coin(self.denom, int(self.amount))

    def to_int_ceil_coin(self) -> Coin:
        """Turns the :class:`coin` into an ``int`` coin with ceiling the amount."""
        return Coin(self.denom, int(math.ceil(self.amount)))

    def to_dec_coin(self) -> Coin:
        """Creates a new :class:`Coin` with a :class:`Dec` amount."""
        return Coin(self.denom, Dec(self.amount))

    def __str__(self) -> str:
        if self.is_dec_coin():
            amount_str = str(self.amount).rstrip("0")
            if amount_str.endswith("."):
                amount_str += "0"
            return f"{amount_str}{self.denom}"
        return f"{self.amount}{self.denom}"

    def to_amino(self) -> dict:
        return {"denom": self.denom, "amount": str(self.amount)}

    def to_data(self) -> dict:
        return {"denom": self.denom, "amount": str(self.amount)}

    @classmethod
    def from_proto(cls, proto: Coin_pb) -> Coin:
        return cls(proto.denom, proto.amount)

    def to_proto(self) -> Coin_pb:
        coin = Coin_pb()
        coin.denom = self.denom
        coin.amount = str(self.amount)
        return coin

    @classmethod
    def from_str(cls, string: str) -> Coin:
        """Creates a new :class:`Coin` from a coin-format string. Must match the format:
        ``283923uluna`` (``int``-Coin) or ``23920.23020uluna`` (:class:`Dec`-Coin).

        >>> int_coin = Coin.from_str("230920uluna")
        >>> int_coin.denom
        'uluna'
        >>> int_coin.amount
        230920
        >>> dec_coin = Coin.from_str("203922.223uluna")
        >>> dec_coin.denom
        'uluna'
        >>> dec_coin.amount
        Dec('203922.223')

        Args:
            string (str): string to convert

        Raises:
            ValueError: if string is in wrong format

        Returns:
            Coin: converted string
        """
        pattern = r"^(\-?[0-9]+(\.[0-9]+)?)([0-9a-zA-Z/]+)$"
        match = re.match(pattern, string)
        if match is None:
            raise ValueError(f"failed to parse Coin: {string}")
        else:
            return cls(match.group(3), match.group(1))

    def add(self, addend: Union[Numeric.Input, Coin]) -> Coin:
        """Creates a new :class:`Coin` with the sum as amount. If the ``addend`` is a
        :class:`Coin`, its ``denom`` must match.

        Args:
            addend (Union[Numeric.Input, Coin]): addend

        Raises:
            ArithmeticError: if addedend has different ``denom``

        Returns:
            Coin: sum
        """
        if isinstance(addend, Coin):
            if addend.denom != self.denom:
                raise ArithmeticError(
                    f"cannot add/subtract two Coin objects of different denoms: {self.denom} and {addend.denom}"
                )
            return Coin(self.denom, self.amount + addend.amount)
        else:
            return Coin(self.denom, self.amount + Numeric.parse(addend))

    def __add__(self, addend: Union[Numeric.Input, Coin]) -> Coin:
        return self.add(addend)

    def sub(self, subtrahend: Union[Numeric.Input, Coin]) -> Coin:
        """Creates a new :class:`Coin` with the difference as amount. If the ``subtrahend`` is a
        :class:`Coin`, its ``denom`` must match.

        Args:
            subtrahend (Union[Numeric.Input, Coin]): subtrahend

        Returns:
            Coin: difference
        """
        if isinstance(subtrahend, Coin):
            return self.add(subtrahend.mul(-1))
        else:
            return self.add(Numeric.parse(subtrahend) * -1)

    def __sub__(self, subtrahend: Union[Numeric.Input, Coin]) -> Coin:
        return self.sub(subtrahend)

    def mul(self, multiplier: Numeric.Input) -> Coin:
        """Creates a new :class:`Coin` with the product as amount. The ``multiplier``
        argument is first coerced to either an ``int`` or :class:`Dec`.

        Args:
            multiplier (Numeric.Input): multiplier

        Returns:
            Coin: product
        """
        other_amount = Numeric.parse(multiplier)
        return Coin(self.denom, self.amount * other_amount)

    def __mul__(self, multiplier: Numeric.Input) -> Coin:
        return self.mul(multiplier)

    def div(self, divisor: Numeric.Input) -> Coin:
        """Creates a new :class:`Coin` with the quotient as amount. The ``divisor``
        argument is first coerced to either an ``int`` or :class:`Dec`.

        Args:
            divisor (Numeric.Input): divisor

        Returns:
            Coin: quotient
        """
        other_amount = Numeric.parse(divisor)
        if isinstance(other_amount, int):
            return Coin(self.denom, (self.amount // other_amount))
        else:
            return Coin(self.denom, (self.amount / other_amount))

    def __truediv__(self, divisor: Numeric.Input) -> Coin:
        return self.div(divisor)

    def __floordiv__(self, divisor: Numeric.Input) -> Coin:
        return self.div(int(Numeric.parse(divisor)))

    def mod(self, modulo: Numeric.Input) -> Coin:
        """Creates a new :class:`Coin` with the modulus as amount.

        Args:
            modulo (Numeric.Input): modulo

        Returns:
            Coin: modulo
        """
        other_amount = Numeric.parse(modulo)
        if isinstance(other_amount, Dec):
            return Coin(self.denom, Dec(self.amount).mod(other_amount))
        else:
            return Coin(self.denom, self.amount % other_amount)

    def __mod__(self, modulo: Numeric.Input) -> Coin:
        return self.mod(modulo)

    def __neg__(self) -> Coin:
        return Coin(denom=self.denom, amount=(-self.amount))

    def __abs__(self) -> Coin:
        return Coin(denom=self.denom, amount=abs(self.amount))

    def __pos__(self) -> Coin:
        return abs(self)

    @classmethod
    def from_data(cls, data: dict) -> Coin:
        """Deserializes a :class:`Coin` object from its JSON data representation.

        Args:
            data (dict): data object
        """
        return cls(data["denom"], data["amount"])

    @classmethod
    def from_amino(cls, data: dict) -> Coin:
        """Deserializes a :class:`Coin` object from its amino-codec representation.

        Args:
            data (dict): data object
        """
        return cls(data["denom"], data["amount"])



from typing import List, Optional, Tuple

from terra_sdk.core import Coins, Dec
from terra_sdk.core.deposit import Deposit
from terra_sdk.core.gov import Proposal, ProposalStatus, WeightedVoteOption
from terra_sdk.core.gov.data import Vote

from ._base import BaseAsyncAPI, sync_bind

__all__ = ["AsyncGovAPI", "GovAPI", "ProposalStatus"]

from ..params import APIParams


class AsyncGovAPI(BaseAsyncAPI):
    async def proposals(
        self, options: dict = {}, params: Optional[APIParams] = None
    ) -> [List[Proposal], dict]:
        """Fetches all proposals.
        Args:
            options (dict, optional): dictionary containing options. Defaults to {}. you can use one or more below:
                {
                    "proposal_status": terra_sdk.core.gov.ProposalStatus (int)
                    "voter": voter address (str),
                    "depositor": depositor address(str)
                }
                example) {"proposal_status":1, "depositor":"terra17lmam6zguazs5q5u6z5mmx76uj63gldnse2pdp"}

            params (APIParams, optional): additional params for the API like pagination

        Returns:
            List[Proposal]: proposals
        """
        if params is not None:
            options.update(params.to_dict())
        res = await self._c._get("/cosmos/gov/v1beta1/proposals", options)
        return [Proposal.from_data(d) for d in res.get("proposals")], res.get(
            "pagination"
        )

    async def proposal(self, proposal_id: int) -> Proposal:
        """Fetches a single proposal by id.

        Args:
            proposal_id (int): proposal ID

        Returns:
            Proposal: proposal
        """
        res = await self._c._get(f"/cosmos/gov/v1beta1/proposals/{proposal_id}")
        return Proposal.from_data(res.get("proposal"))

    # keep it private
    async def __search_submit_proposal(self, proposal_id: int):
        params = [
            ("message.action", "/cosmos.gov.v1beta1.MsgSubmitProposal"),
            ("submit_proposal.proposal_id", proposal_id),
        ]

        res = await self._c._search(params)
        txs = res.get("txs")
        if txs is None or len(txs) <= 0:
            raise Exception("failed to find submit proposal")
        return txs[0]

    # keep it private
    async def __search_deposits(
        self, proposal_id: int, params: Optional[APIParams] = None
    ):
        events = [
            ("message.action", "/cosmos.gov.v1beta1.MsgDeposit"),
            ("proposal_deposit.proposal_id", proposal_id),
        ]
        if params is not None:
            d = params.to_dict()
            for i in d.keys():
                events.append((i, d.get(i)))
        res = await self._c._search(events)
        txs = res.get("txs")
        if txs is None or len(txs) <= 0:
            raise Exception("failed to find deposit txs")
        return txs, res.get("pagination")

    # keep it private
    async def __search_votes(
        self, proposal_id: int, action: str, params: Optional[APIParams] = None
    ):
        events = [
            ("message.action", "/cosmos.gov.v1beta1.MsgVote"),
            ("proposal_vote.proposal_id", proposal_id),
        ]
        if params is not None:
            d = params.to_dict()
            for i in d.keys():
                events.append((i, d.get(i)))

        res = await self._c._search(events)
        txs = res.get("txs")
        if txs is None or len(txs) <= 0:
            raise Exception("failed to find vote txs")
        return txs, res.get("pagination")

    async def proposer(self, proposal_id: int) -> str:
        """Fetches the proposer of a proposal.

        Args:
            proposal_id (int): proposal ID

        Returns:
            str: proposal's proposer, None if proposal is not exist
        """

        res = await self.__search_submit_proposal(proposal_id)
        msgs = res["body"]["messages"]
        for msg in msgs:
            if msg.get("@type") == "/cosmos.gov.v1beta1.MsgSubmitProposal":
                return msg["proposer"]
        return None

    async def deposits(self, proposal_id: int, params: Optional[APIParams] = None):
        """Fetches the deposit information about a proposal.

        Args:
            proposal_id (int): proposal ID
            params (APIParams, optional): additional params for the API like pagination
        """

        proposal = self.proposal(proposal_id)

        status = proposal.status
        if (
            status == ProposalStatus.PROPOSAL_STATUS_DEPOSIT_PERIOD.name
            or status == ProposalStatus.PROPOSAL_STATUS_VOTING_PERIOD.name
        ):
            res = await self._c._get(
                f"/cosmos/gov/v1beta1/proposals/{proposal_id}/deposits", params
            )
            return [Deposit.from_data(d) for d in res.get("deposits")]

        res, pagination = await self.__search_deposits(proposal_id, params)
        deposits = []
        for tx in res:
            for msg in tx.get("body").get("messages"):
                if msg.get("@type") == "/cosmos.gov.v1beta1.MsgDeposit":
                    deposits.append(Deposit.from_data(msg))
        return deposits, pagination

    async def votes(self, proposal_id: int, params: Optional[APIParams] = None):
        """Fetches the votes for a proposal.

        Args:
            proposal_id (int): proposal ID
            params (APIParams, optional): additional params for the API like pagination
        """

        proposal = self.proposal(proposal_id)
        if proposal.status == ProposalStatus.PROPOSAL_STATUS_DEPOSIT_PERIOD:
            res = await self._c._get(
                f"/cosmos/gov/v1beta1/proposals/{proposal_id}/votes", params
            )
            return res.get("votes"), res.get("pagination")

        res, pagination = await self.__search_votes(proposal_id, params)
        votes = []
        for tx in res:
            for msg in tx.get("body").get("messages"):
                if (
                    msg.get("@type") == "/cosmos.gov.v1beta1.MsgVote"
                    and msg.get("proposal_id") == proposal_id
                ):
                    votes.append(WeightedVoteOption(msg.get("option"), 1))
                elif (
                    msg.get("@type") == "/cosmos.gov.v1beta1.MsgVoteWeighted"
                    and msg.get("proposal_id") == proposal_id
                ):
                    votes.append(
                        Vote(
                            proposal_id=proposal_id,
                            voter=msg.get("voter"),
                            options=msg.get("options"),
                        )
                    )
        return votes, pagination

    async def tally(self, proposal_id: int):
        """Fetches the tally for a proposal.

        Args:
            proposal_id (int): proposal ID
        """
        res = await self._c._get(f"/cosmos/gov/v1beta1/proposals/{proposal_id}/tally")
        return res.get("tally")

    async def deposit_parameters(self) -> dict:
        """Fetches the Gov module's deposit parameters.

        Returns:
            dict: deposit parameters
        """
        result = await self._c._get("/cosmos/gov/v1beta1/params/deposit")
        params = result.get("deposit_params")
        return {
            "max_deposit_period": params["max_deposit_period"],
            "min_deposit": Coins.from_data(params["min_deposit"]),
        }

    async def voting_parameters(self) -> dict:
        """Fetches the Gov module's voting parameters.

        Returns:
            dict: voting parameters
        """
        result = await self._c._get("/cosmos/gov/v1beta1/params/voting")
        return result.get("voting_params")

    async def tally_parameters(self) -> dict:
        """Fetches the Gov module's tally parameters.

        Returns:
            dict: tally parameters
        """
        result = await self._c._get("/cosmos/gov/v1beta1/params/tallying")
        params = result.get("tally_params")
        return {
            "quorum": Dec(params["quorum"]),
            "threshold": Dec(params["threshold"]),
            "veto_threshold": Dec(params["veto_threshold"]),
        }

    async def parameters(self) -> dict:
        """Fetches the Gov module's parameters.

        Returns:
            dict: Gov module parameters
        """
        return {
            "deposit_params": await BaseAsyncAPI._try_await(self.deposit_parameters()),
            "voting_params": await BaseAsyncAPI._try_await(self.voting_parameters()),
            "tally_params": await BaseAsyncAPI._try_await(self.tally_parameters()),
        }


class GovAPI(AsyncGovAPI):
    @sync_bind(AsyncGovAPI.proposals)
    def proposals(self, params: Optional[APIParams] = None) -> Tuple[List[Proposal], dict]:
        pass

    proposals.__doc__ = AsyncGovAPI.proposals.__doc__

    @sync_bind(AsyncGovAPI.proposal)
    def proposal(self, proposal_id: int) -> Proposal:
        pass

    proposal.__doc__ = AsyncGovAPI.proposal.__doc__

    @sync_bind(AsyncGovAPI.proposer)
    def proposer(self, proposal_id: int) -> str:
        pass

    proposer.__doc__ = AsyncGovAPI.proposer.__doc__

    @sync_bind(AsyncGovAPI.deposits)
    def deposits(self, proposal_id: int, params: Optional[APIParams] = None):
        pass

    deposits.__doc__ = AsyncGovAPI.deposits.__doc__

    @sync_bind(AsyncGovAPI.votes)
    def votes(self, proposal_id: int, params: Optional[APIParams] = None):
        pass

    votes.__doc__ = AsyncGovAPI.votes.__doc__

    @sync_bind(AsyncGovAPI.tally)
    def tally(self, proposal_id: int):
        pass

    tally.__doc__ = AsyncGovAPI.tally.__doc__

    @sync_bind(AsyncGovAPI.deposit_parameters)
    def deposit_parameters(self) -> dict:
        pass

    deposits.__doc__ = AsyncGovAPI.deposit_parameters.__doc__

    @sync_bind(AsyncGovAPI.voting_parameters)
    def voting_parameters(self) -> dict:
        pass

    voting_parameters.__doc__ = AsyncGovAPI.voting_parameters.__doc__

    @sync_bind(AsyncGovAPI.tally_parameters)
    def tally_parameters(self) -> dict:
        pass

    tally_parameters.__doc__ = AsyncGovAPI.tally_parameters.__doc__

    @sync_bind(AsyncGovAPI.parameters)
    def parameters(self) -> dict:
        pass

    parameters.__doc__ = AsyncGovAPI.parameters.__doc__



"""wasm module governance proposal types."""

from __future__ import annotations

import base64
import json
from copyreg import add_extension
from curses import init_color
from typing import List, Union

import attr
from betterproto.lib.google.protobuf import Any as Any_pb
from terra_proto.cosmwasm.wasm.v1 import ClearAdminProposal as ClearAdminProposal_pb
from terra_proto.cosmwasm.wasm.v1 import (
    ExecuteContractProposal as ExecuteContractProposal_pb,
)
from terra_proto.cosmwasm.wasm.v1 import (
    InstantiateContractProposal as InstantiateContractProposal_pb,
)
from terra_proto.cosmwasm.wasm.v1 import (
    MigrateContractProposal as MigrateContractProposal_pb,
)
from terra_proto.cosmwasm.wasm.v1 import PinCodesProposal as PinCodesProposal_pb
from terra_proto.cosmwasm.wasm.v1 import StoreCodeProposal as StoreCodeProposal_pb
from terra_proto.cosmwasm.wasm.v1 import SudoContractProposal as SudoContractProposal_pb
from terra_proto.cosmwasm.wasm.v1 import UnpinCodesProposal as UnpinCodesProposal_pb
from terra_proto.cosmwasm.wasm.v1 import UpdateAdminProposal as UpdateAdminProposal_pb
from terra_proto.cosmwasm.wasm.v1 import (
    UpdateInstantiateConfigProposal as UpdateInstantiateConfigProposal_pb,
)

from terra_sdk.core.bech32 import AccAddress
from terra_sdk.core.coins import Coins
from terra_sdk.core.wasm.data import AccessConfig, AccessConfigUpdate
from terra_sdk.util.json import JSONSerializable
from terra_sdk.util.remove_none import remove_none

__all__ = [
    "ClearAdminProposal",
    "ExecuteContractProposal",
    "InstantiateContractProposal",
    "MigrateContractProposal",
    "PinCodesProposal",
    "StoreCodeProposal",
    "SudoContractProposal",
    "UnpinCodesProposal",
    "UpdateAdminProposal",
    "UpdateInstantiateConfigProposal",
]


def parse_msg(msg: Union[dict, str, bytes]) -> dict:
    if type(msg) is dict:
        return msg
    return json.loads(msg)


@attr.s
class ClearAdminProposal(JSONSerializable):
    """
    Args:
        title : a short summary
        description : a human readable text
        contract : the address of the smart contract
    """

    type_amino = "wasm/ClearAdminProposal"
    type_url = "/cosmwasm.wasm.v1.ClearAdminProposal"

    title: str = attr.ib()
    description: str = attr.ib()
    contract: AccAddress = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "title": self.title,
                "description": self.description,
                "contract": self.contract,
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "title": self.title,
            "description": self.description,
            "contract": self.contract,
        }

    def to_proto(self) -> ClearAdminProposal:
        return ClearAdminProposal_pb(
            title=self.title, description=self.description, contract=self.contract
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    @classmethod
    def from_data(cls, data):
        return cls(
            title=data["title"],
            description=data["description"],
            contract=data["contract"],
        )

    @classmethod
    def from_proto(cls, proto: ClearAdminProposal_pb):
        return cls(
            title=proto.title, description=proto.description, contract=proto.contract
        )


@attr.s
class ExecuteContractProposal(JSONSerializable):
    """
    Args:
        title : a short summary
        description : a human readable text
        run_as : contract user
        contract : the address of the smart contract
        execute_msg : HandleMsg to pass as arguments for contract invocation
        coins : coins to be sent to contract
    """

    type_amino = "wasm/ExecuteContractProposal"
    type_url = "/cosmwasm.wasm.v1.ExecuteContractProposal"
    prototype = ExecuteContractProposal_pb

    title: str = attr.ib()
    description: str = attr.ib()
    run_as: AccAddress = attr.ib()
    contract: AccAddress = attr.ib()
    execute_msg: Union[dict, str] = attr.ib()
    coins: Coins = attr.ib(converter=Coins, factory=Coins)

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "title": self.title,
                "description": self.description,
                "run_as": self.run_as,
                "contract": self.contract,
                "execute_msg": remove_none(self.execute_msg),
                "coins": self.coins.to_amino(),
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "title": self.title,
            "description": self.description,
            "run_as": self.run_as,
            "contract": self.contract,
            "execute_msg": remove_none(self.execute_msg),
            "coins": self.coins.to_data(),
        }

    def to_proto(self) -> ExecuteContractProposal:
        return ExecuteContractProposal_pb(
            title=self.title,
            description=self.description,
            run_as=self.run_as,
            contract=self.contract,
            execute_msg=bytes(json.dumps(self.execute_msg), "utf-8"),
            coins=self.coins.to_proto(),
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    @classmethod
    def from_data(cls, data):
        return cls(
            title=data["title"],
            description=data["description"],
            run_as=data["run_as"],
            contract=data["contract"],
            execute_msg=parse_msg(data["execute_msg"]),
            coins=Coins.from_data(data["coins"]),
        )

    @classmethod
    def from_proto(cls, proto: ExecuteContractProposal_pb):
        return cls(
            title=proto.title,
            description=proto.description,
            run_as=proto.run_as,
            contract=proto.contract,
            execute_msg=parse_msg(proto.msg),
            coins=Coins.from_proto(proto.funds),
        )


@attr.s
class InstantiateContractProposal(JSONSerializable):
    """
    Args:
        title : a short summary
        description : a human readable text
        run_as : contract user
        admin : an optional contract admin address who can migrate the contract, put empty string to disable migration
        code_id : the reference to the stored WASM code
        init_msg : json encoded message to be passed to the contract on instantiation
        init_coins : transferred to the contract on execution
        label : label for the contract. v2 supported only
    """

    type_amino = "wasm/InstantiateContractProposal"
    type_url = "/cosmwasm.wasm.v1.InstantiateContractProposal"
    prototype = InstantiateContractProposal_pb

    title: str = attr.ib()
    description: str = attr.ib()
    run_as: AccAddress = attr.ib()
    admin: Union[AccAddress, None] = attr.ib()
    code_id: int = attr.ib()
    init_msg: Union[dict, str] = attr.ib()
    init_coins: Coins = attr.ib(converter=Coins)
    label: str = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "title": self.title,
                "description": self.description,
                "run_as": self.run_as,
                "admin": self.admin,
                "code_id": self.code_id,
                "init_msg": remove_none(self.init_msg),
                "init_coins": self.init_coins.to_amino(),
                "label": self.label,
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "title": self.title,
            "description": self.description,
            "run_as": self.run_as,
            "admin": self.admin,
            "code_id": self.code_id,
            "init_msg": remove_none(self.init_msg),
            "init_coins": self.init_coins.to_amino(),
            "label": self.label,
        }

    def to_proto(self) -> InstantiateContractProposal:
        return InstantiateContractProposal_pb(
            title=self.title,
            description=self.description,
            run_as=self.run_as,
            admin=self.admin,
            code_id=self.code_id,
            init_msg=bytes(json.dumps(self.init_msg), "utf-8"),
            init_coins=self.init_coins.to_proto(),
            label=self.label,
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    @classmethod
    def from_data(cls, data):
        return cls(
            title=data["title"],
            description=data["description"],
            run_as=data["run_as"],
            admin=data["admin"],
            code_id=data["code_id"],
            init_msg=parse_msg(data["init_msg"]),
            init_coins=Coins.from_data(data["init_coins"]),
            label=data["label"],
        )

    @classmethod
    def from_proto(cls, proto: InstantiateContractProposal_pb):
        return cls(
            title=proto.title,
            description=proto.description,
            run_as=proto.run_as,
            admin=proto.admin,
            code_id=proto.code_id,
            init_msg=parse_msg(proto.msg),
            init_coins=Coins.from_proto(proto.funds),
            label=proto.label,
        )


@attr.s
class MigrateContractProposal(JSONSerializable):
    """
    Args:
        title : a short summary
        description : a human readable text
        contract : contract address to be migrated from
        new_code_id : reference to the new code on the blockchain
        migrate_msg : JSON message to configure the migrate state of the contract
    """

    type_amino = "wasm/MigrateContractProposal"
    type_url = "/cosmwasm.wasm.v1.MigrateContractProposal"
    prototype = MigrateContractProposal_pb

    title: str = attr.ib()
    description: str = attr.ib()
    contract: AccAddress = attr.ib()
    new_code_id: int = attr.ib()
    migrate_msg: Union[dict, str] = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "title": self.title,
                "description": self.description,
                "contract": self.contract,
                "new_code_id": self.new_code_id,
                "migrate_msg": remove_none(self.migrate_msg),
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "title": self.title,
            "description": self.description,
            "contract": self.contract,
            "new_code_id": self.new_code_id,
            "migrate_msg": remove_none(self.migrate_msg),
        }

    def to_proto(self) -> MigrateContractProposal:
        return MigrateContractProposal_pb(
            title=self.title,
            description=self.description,
            contract=self.contract,
            new_code_id=self.new_code_id,
            migrate_msg=bytes(json.dumps(self.migrate_msg), "utf-8"),
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    @classmethod
    def from_data(cls, data):
        return cls(
            title=data["title"],
            description=data["description"],
            contract=data["contract"],
            new_code_id=data["new_code_id"],
            migrate_msg=parse_msg(data["migrate_msg"]),
        )

    @classmethod
    def from_proto(cls, proto: MigrateContractProposal_pb):
        return cls(
            title=proto.title,
            description=proto.description,
            contract=proto.contract,
            new_code_id=proto.code_id,
            migrate_msg=parse_msg(proto.msg),
        )


@attr.s
class PinCodesProposal(JSONSerializable):
    """
    Args:
        title : a short summary
        description : a human readable text
        code_ids : the address of the smart code_ids
    """

    type_amino = "wasm/PinCodesProposal"
    type_url = "/cosmwasm.wasm.v1.PinCodesProposal"

    title: str = attr.ib()
    description: str = attr.ib()
    code_ids: List[int] = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "title": self.title,
                "description": self.description,
                "code_ids": [code_id for code_id in self.code_ids],
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "title": self.title,
            "description": self.description,
            "code_ids": [code_id for code_id in self.code_ids],
        }

    def to_proto(self) -> PinCodesProposal:
        return PinCodesProposal_pb(
            title=self.title,
            description=self.description,
            code_ids=[code_id for code_id in self.code_ids],
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    @classmethod
    def from_data(cls, data):
        return cls(
            title=data["title"],
            description=data["description"],
            code_ids=[code_id for code_id in data["code_ids"]],
        )

    @classmethod
    def from_proto(cls, proto: PinCodesProposal_pb):
        return cls(
            title=proto.title,
            description=proto.description,
            code_ids=[code_id for code_id in proto.code_ids],
        )


@attr.s
class StoreCodeProposal(JSONSerializable):
    """
    Args:
        title : a short summary
        description : a human readable text
        run_as : the address that is passed to the contract's environment as sender
        wasm_byte_code : can be raw or gzip compressed
        instantiate_permission : to apply on contract creation, optional
    """

    type_amino = "wasm/StoreCodeProposal"
    type_url = "/cosmwasm.wasm.v1.StoreCodeProposal"

    title: str = attr.ib()
    description: str = attr.ib()
    run_as: str = attr.ib()
    wasm_byte_code: str = attr.ib()
    instantiate_permission: AccessConfig = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "title": self.title,
                "description": self.description,
                "run_as": self.run_as,
                "wasm_byte_code": self.wasm_byte_code,
                "instantiate_permission": self.instantiate_permission.to_amino(),
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "title": self.title,
            "description": self.description,
            "run_as": self.run_as,
            "wasm_byte_code": self.wasm_byte_code,
            "instantiate_permission": self.instantiate_permission.to_data(),
        }

    def to_proto(self) -> StoreCodeProposal:
        return StoreCodeProposal_pb(
            title=self.title,
            description=self.description,
            run_as=self.run_as,
            wasm_byte_code=base64.b64encode(self.wasm_byte_code),
            instantiate_permission=self.instantiate_permission,
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    @classmethod
    def from_data(cls, data):
        return cls(
            title=data["title"],
            description=data["description"],
            run_as=data["run_as"],
            wasm_byte_code=data["wasm_byte_code"],
            instantiate_permission=AccessConfig.from_data(
                data["instantiate_permission"]
            ),
        )

    @classmethod
    def from_proto(cls, proto: StoreCodeProposal_pb):
        return cls(
            title=proto.title,
            description=proto.description,
            run_as=proto.run_as,
            wasm_byte_code=base64.b64encode(proto.wasm_byte_code).decode(),
            instantiate_permission=AccessConfig.from_proto(
                proto.instantiate_permission
            ),
        )


@attr.s
class SudoContractProposal(JSONSerializable):
    """
    Args:
        title : a short summary
        description : a human readable text
        contract :  contract address to be migrated from
        msg : JSON message to configure the migrate state of the contract
    """

    type_amino = "wasm/SudoContractProposal"
    type_url = "/cosmwasm.wasm.v1.SudoContractProposal"

    title: str = attr.ib()
    description: str = attr.ib()
    contract: AccAddress = attr.ib()
    msg: Union[dict, str] = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "title": self.title,
                "description": self.description,
                "contract": self.contract,
                "msg": remove_none(self.msg),
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "title": self.title,
            "description": self.description,
            "contract": self.contract,
            "msg": remove_none(self.msg),
        }

    def to_proto(self) -> SudoContractProposal:
        return SudoContractProposal_pb(
            title=self.title,
            description=self.description,
            contract=self.contract,
            msg=bytes(json.dumps(self.msg), "utf-8"),
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    @classmethod
    def from_data(cls, data):
        return cls(
            title=data["title"],
            description=data["description"],
            contract=data["contract"],
            msg=parse_msg(data["msg"]),
        )

    @classmethod
    def from_proto(cls, proto: SudoContractProposal_pb):
        return cls(
            title=proto.title,
            description=proto.description,
            contract=proto.contract,
            msg=parse_msg(proto.msg),
        )


@attr.s
class UnpinCodesProposal(JSONSerializable):
    """
    Args:
        title : a short summary
        description : a human readable text
        code_ids : the address of the smart code_ids
    """

    type_amino = "wasm/UnpinCodesProposal"
    type_url = "/cosmwasm.wasm.v1.UnpinCodesProposal"

    title: str = attr.ib()
    description: str = attr.ib()
    code_ids: List[int] = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "title": self.title,
                "description": self.description,
                "code_ids": [code_id for code_id in self.code_ids],
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "title": self.title,
            "description": self.description,
            "code_ids": [code_id for code_id in self.code_ids],
        }

    def to_proto(self) -> UnpinCodesProposal:
        return UnpinCodesProposal_pb(
            title=self.title,
            description=self.description,
            code_ids=[code_id for code_id in self.code_ids],
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    @classmethod
    def from_data(cls, data):
        return cls(
            title=data["title"],
            description=data["description"],
            code_ids=[code_id for code_id in data["code_ids"]],
        )

    @classmethod
    def from_proto(cls, proto: UnpinCodesProposal_pb):
        return cls(
            title=proto.title,
            description=proto.description,
            code_ids=[code_id for code_id in proto.code_ids],
        )


@attr.s
class UpdateAdminProposal(JSONSerializable):
    """
    Args:
        title : a short summary
        description : a human readable text
        contract : the address of the smart contract
        new_admin : address to be set
    """

    type_amino = "wasm/UpdateAdminProposal"
    type_url = "/cosmwasm.wasm.v1.UpdateAdminProposal"

    title: str = attr.ib()
    description: str = attr.ib()
    contract: AccAddress = attr.ib()
    new_admin: AccAddress = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "title": self.title,
                "description": self.description,
                "contract": self.contract,
                "new_admin": self.new_admin,
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "title": self.title,
            "description": self.description,
            "contract": self.contract,
            "new_admin": self.new_admin,
        }

    def to_proto(self) -> UpdateAdminProposal:
        return UpdateAdminProposal_pb(
            title=self.title,
            description=self.description,
            contract=self.contract,
            new_admin=self.new_admin,
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    @classmethod
    def from_data(cls, data):
        return cls(
            title=data["title"],
            description=data["description"],
            contract=data["contract"],
            new_admin=data["new_admin"],
        )

    @classmethod
    def from_proto(cls, proto: UpdateAdminProposal_pb):
        return cls(
            title=proto.title,
            description=proto.description,
            contract=proto.contract,
            new_admin=proto.new_admin,
        )


@attr.s
class UpdateInstantiateConfigProposal(JSONSerializable):
    """
    Args:
        title : a short summary
        description : a human readable text
        access_config_updates : the address of the smart access_config_updates
    """

    type_amino = "wasm/UpdateInstantiateConfigProposal"
    type_url = "/cosmwasm.wasm.v1.UpdateInstantiateConfigProposal"

    title: str = attr.ib()
    description: str = attr.ib()
    access_config_updates: List[AccessConfigUpdate] = attr.ib()

    def to_amino(self) -> dict:
        return {
            "type": self.type_amino,
            "value": {
                "title": self.title,
                "description": self.description,
                "access_config_updates": [
                    access_config_update.to_amino()
                    for access_config_update in self.access_config_updates
                ],
            },
        }

    def to_data(self) -> dict:
        return {
            "@type": self.type_url,
            "title": self.title,
            "description": self.description,
            "access_config_updates": [
                access_config_update.to_data()
                for access_config_update in self.access_config_updates
            ],
        }

    def to_proto(self) -> UpdateInstantiateConfigProposal:
        return UpdateInstantiateConfigProposal_pb(
            title=self.title,
            description=self.description,
            access_config_updates=[
                access_config_update.to_proto()
                for access_config_update in self.access_config_updates
            ],
        )

    def pack_any(self) -> Any_pb:
        return Any_pb(type_url=self.type_url, value=bytes(self.to_proto()))

    @classmethod
    def from_data(cls, data):
        return cls(
            title=data["title"],
            description=data["description"],
            access_config_updates=[
                AccessConfigUpdate.from_data(access_config_update)
                for access_config_update in data["access_config_updates"]
            ],
        )

    @classmethod
    def from_proto(cls, proto: UpdateInstantiateConfigProposal_pb):
        return cls(
            title=proto.title,
            description=proto.description,
            access_config_updates=[
                AccessConfigUpdate.from_proto(access_config_update)
                for access_config_update in proto.access_config_updates
            ],
        )
